from __future__ import annotations

from typing import Optional

import numpy as np
import pandas as pd
from seeq.sdk import *
from seeq.spy import _common
from seeq.spy import _login
from seeq.spy._errors import *
from seeq.spy._redaction import safely, request_safely
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.workbooks._folder import SYNTHETIC_FOLDERS, synthetic_folder_to_content_filter


def search(query, *, content_filter='owner', all_properties=False, recursive=False, include_archived=False,
           errors=None, quiet=False, status=None, session: Optional[Session] = None):
    """
    Issues a query to the Seeq Server to retrieve metadata for workbooks.
    This metadata can be used to pull workbook definitions into memory.

    Parameters
    ----------
    query : dict
        A mapping of property / match-criteria pairs. Match criteria uses
        the same syntax as the Data tab in Seeq (contains, or glob, or regex).
        Available options are:

        =================== ===================================================
        Property            Description
        =================== ===================================================
        ID                  ID of the workbook, as seen in the URL.
        Name                Name of the workbook.
        Path                Path to the workbook through the folder hierarchy.
        Description         Description of the workbook.
        Workbook Type       Either 'Analysis' or 'Topic'.
        =================== ===================================================

    content_filter : str, default 'owner'
        The allowed_content_filter must be updated every time there is a change to the ItemWithOwner#Filter
        Filters workbooks according to the following possible values:

        =================== ===================================================
        Property            Description
        =================== ===================================================
        owner               Only content owned by the logged-in user.
        shared              Only content shared specifically with the logged-in
                            user. Note: To obtain the same results as you would
                            when looking at the "Shared" tab in the Seeq home
                            screen, use "sharedorpublic" (see below).
        public              Only content shared with Everyone.
        sharedorpublic      Content shared with Everyone or with the logged-in user
        corporate           Only content in the Corporate folder
        all                 All content, across all users (logged-in user must
                            be admin).
        users               Every user's home folder (logged-in user must be admin)
        =================== ===================================================

    all_properties : bool, default False
        True if all workbook properties should be retrieved. This currently makes
        the search operation much slower as retrieval of properties for an item
        requires a separate call.

    recursive : bool, default False
        True if workbooks further down in the folder path should be returned.

    include_archived : bool, default False
        True if archived (aka "trashed") folders/workbooks should be returned.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If 'catalog',
        errors will be added to a 'Result' column in the status.df DataFrame.

    quiet : bool, default False
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedent.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with rows for each workbook found and columns for each
        property.

    """
    _common.validate_argument_types([
        (query, 'query', dict),
        (content_filter, 'content_filter', str),
        (all_properties, 'all_properties', bool),
        (recursive, 'recursive', bool),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    status = Status.validate(status, quiet, errors)
    session = Session.validate(session)
    _login.validate_login(session, status)

    status.reset_timer()

    try:
        results_df = _search(session, query, content_filter=content_filter, all_properties=all_properties,
                             recursive=recursive, include_archived=include_archived, status=status)

    except KeyboardInterrupt:
        status.update('Query canceled.', Status.CANCELED)
        return

    status.metrics({
        'Results': {
            'Time': status.get_timer(),
            'Count': len(results_df)
        }
    })
    status.update('Query successful.', Status.SUCCESS)

    return results_df


def _search(session: Session, query, status: Status, *, content_filter, all_properties, recursive, include_archived,
            parent_id=None, parent_path='', search_folder_id=None) -> pd.DataFrame:
    items_api = ItemsApi(session.client)
    workbooks_api = WorkbooksApi(session.client)
    results_df = pd.DataFrame()

    content_filter = content_filter.upper()
    allowed_content_filters = ['OWNER', 'SHARED', 'PUBLIC', 'CORPORATE', 'ALL', 'USERS', 'SHAREDORPUBLIC']
    if content_filter not in allowed_content_filters:
        raise SPyValueError('content_filter must be one of: %s' % ', '.join(allowed_content_filters))

    for _key, _ in query.items():
        supported_query_fields = ['ID', 'Path', 'Name', 'Description', 'Workbook Type']
        if _key not in supported_query_fields:
            raise SPyValueError('"%s" unsupported query field, use instead one or more of: %s' %
                                (_key, ', '.join(supported_query_fields)))

    if 'ID' in query:
        workbook_output: Optional[WorkbookOutputV1] = safely(
            lambda: workbooks_api.get_workbook(id=query['ID']),
            action_description=f'get the details for Workbook {query["ID"]}',
            status=status)
        if workbook_output is None:
            return pd.DataFrame()

        content_dict = {
            'ID': workbook_output.id,
            'Type': 'Workbook',
            'Workbook Type': workbook_output.type,
            'Path': _common.path_list_to_string([a.name for a in workbook_output.ancestors]),
            'Name': workbook_output.name,
            'Archived': workbook_output.is_archived,
            'Pinned': workbook_output.marked_as_favorite,
            'Created At': pd.to_datetime(workbook_output.created_at),
            'Updated At': pd.to_datetime(workbook_output.updated_at)
        }
        if workbook_output.owner:
            content_dict['Owner Name'] = workbook_output.owner.name
            content_dict['Owner Username'] = workbook_output.owner.username
            content_dict['Owner ID'] = workbook_output.owner.id

        return pd.DataFrame([content_dict])

    path_filter = query['Path'] if 'Path' in query else None

    path_filter_parts = list()
    path_start = None

    if path_filter is not None:
        path_filter_parts = _common.path_string_to_list(path_filter)
        if len(path_filter_parts) > 0:
            path_start = path_filter_parts[0]

    # We want the predefined synthetic folders like workbooks.CORPORATE and workbooks.MY_FOLDER to
    # work as root directories when specified at the beginning of Path
    if path_start in SYNTHETIC_FOLDERS:
        path_start_no_underscore = path_start.replace('__', '').replace('_', ' ')
        warning_message = f'The content_filter {content_filter} was overwritten, and the searched directory was ' \
                          f'set to {path_start_no_underscore} based on the provided path.'

        new_content_filter = synthetic_folder_to_content_filter(path_start)

        if content_filter != new_content_filter and status is not None:
            status.warn(warning_message)
        # Apply the corresponding content_filter and use the rest of the path
        content_filter = new_content_filter
        path_filter_parts.pop(0)
    contents = list()

    _folder_id_sub_description = f'within {parent_id} ' if parent_id else ''
    _request_folder_contents_description = f'get Folders {_folder_id_sub_description}using filter {content_filter}'

    @request_safely(action_description=_request_folder_contents_description,
                    additional_errors=[400], status=status)
    def _add_to_folder_contents(archived):
        folder_output_list: WorkbenchItemOutputListV1 = get_folders(session,
                                                                    content_filter=content_filter,
                                                                    folder_id=parent_id,
                                                                    archived=archived)

        contents.extend(folder_output_list.content)

    _add_to_folder_contents(False)
    if include_archived:
        _add_to_folder_contents(True)

    for content in contents:  # type: WorkbenchSearchResultPreviewV1
        path_matches = False
        props_match = True
        if content.type == 'Folder' and len(path_filter_parts) > 0 and \
                _common.does_query_fragment_match(path_filter_parts[0], content.name, contains=False):
            path_matches = True

        for query_key, content_key in [('Name', 'name'), ('Description', 'description')]:
            attr_value = getattr(content, content_key)
            if query_key in query and (attr_value is None or
                                       not _common.does_query_fragment_match(query[query_key], attr_value)):
                props_match = False
                break

        workbook_type = content.type

        if 'Workbook Type' in query and not _common.does_query_fragment_match(query['Workbook Type'], workbook_type):
            props_match = False

        absolute_path = parent_path

        _type = content.type or np.nan
        if _type in ['Analysis', 'Topic']:
            # This is for backward compatibility with .49 and earlier, which used the same type (Workbook) for both
            # Analysis and Topic. Eventually we may want to deprecate "Workbook Type" and fold it into the "Type"
            # property.
            _type = 'Workbook'

        if props_match and len(path_filter_parts) == 0:
            content_dict = {
                'ID': content.id or np.nan,
                'Type': _type,
                'Path': absolute_path,
                'Name': content.name or np.nan,
                'Pinned': content.is_pinned or np.nan,
                'Archived': content.is_archived or np.nan,
                'Created At': pd.to_datetime(content.created_at),
                'Updated At': pd.to_datetime(content.updated_at)
            }

            if search_folder_id:
                content_dict['Search Folder ID'] = search_folder_id

            content_dict['Workbook Type'] = workbook_type

            if all_properties:
                excluded_properties = [
                    # Exclude these because they're in ns-since-epoch when we retrieve them this way
                    'Created At', 'Updated At',

                    # Exclude this because it's a bunch of JSON that will clutter up the DataFrame
                    'Data', 'workbookState'
                ]

                def _add_error_message_and_warn(msg):
                    content_dict['Pull Result'] = msg
                    status.warn(msg)

                @request_safely(action_description=f'get all properties for Workbook "{content.name}" {content.id}',
                                status=status, on_error=_add_error_message_and_warn)
                def _request_workbook_properties():
                    _item: ItemOutputV1 = items_api.get_item_and_all_properties(id=content.id)
                    for prop in _item.properties:  # type: PropertyOutputV1
                        if prop.name not in excluded_properties:
                            content_dict[prop.name] = _common.none_to_nan(prop.value)

                _request_workbook_properties()

            results_df = results_df.append(content_dict, ignore_index=True, sort=True)

        if content.type == 'Folder' and ((recursive and len(path_filter_parts) == 0) or path_matches):
            child_path_filter = None
            if path_filter_parts and len(path_filter_parts) > 1:
                child_path_filter = _common.path_list_to_string(path_filter_parts[1:])

            if len(parent_path) == 0:
                new_parent_path = content.name
            else:
                new_parent_path = parent_path + ' >> ' + content.name

            child_query = dict(query)
            if not child_path_filter and 'Path' in child_query:
                # We've finished drilling down using the provided 'Path' so now we can use the current folder ID as the
                # "root" from which all paths can be made relative (if desired)
                search_folder_id = content.id
                del child_query['Path']
            else:
                child_query['Path'] = child_path_filter

            child_results_df = _search(session, child_query, status, content_filter=content_filter,
                                       all_properties=all_properties, recursive=recursive,
                                       include_archived=include_archived, parent_id=content.id,
                                       parent_path=new_parent_path, search_folder_id=search_folder_id)

            results_df = results_df.append(child_results_df, ignore_index=True, sort=True)

    return results_df


def get_folders(session: Session, content_filter='ALL', folder_id=None, archived=False, sort_order='createdAt ASC',
                only_pinned=False) -> WorkbenchItemOutputListV1:
    folders_api = FoldersApi(session.client)

    kwargs = {
        'is_archived': archived,
        'sort_order': sort_order,
        'limit': 100000,
        'only_pinned': only_pinned
    }

    if _common.is_guid(folder_id):
        kwargs['folder_id'] = folder_id
    elif content_filter in ('OWNER', 'CORPORATE', 'USERS'):
        kwargs['folder_id'] = content_filter
    else:
        kwargs['filter'] = content_filter

    folders = folders_api.get_folders(**kwargs)

    return folders



from __future__ import annotations

import pandas as pd

reference_types = ['StoredSignal', 'StoredCondition']
calculated_types = ['CalculatedScalar', 'CalculatedSignal', 'CalculatedCondition']
metric_types = ['Metric', 'ThresholdMetric']
data_types = calculated_types + reference_types
supported_input_types = data_types + metric_types + ['Asset']
supported_output_types = calculated_types + ['Asset']

dataframe_dtypes = {
    'ID': str,
    'Referenced ID': str,
    'Path': str,
    'Name': str,
    'Type': str,
    'Depth': int,
    'Description': str,
    'Formula': str,
    'Formula Parameters': (str, list, dict, pd.Series, pd.DataFrame),
    'Roll Up Statistic': str,
    'Roll Up Parameters': str,
    # Below are metric-specific columns
    'Aggregation Function': str,
    'Statistic': str,  # 'Statistic' is the friendly SPy input for 'Aggregation Function'
    'Bounding Condition': (str, dict),
    'Bounding Condition Maximum Duration': str,
    'Duration': str,
    'Number Format': str,
    'Measured Item': (str, dict),
    'Metric Neutral Color': str,
    'Period': str,
    'Process Type': str,
    'Thresholds': (dict, list),
    'Template ID': str
}
dataframe_columns = list(dataframe_dtypes.keys())

MAX_ERRORS_DISPLAYED = 3
MAX_FORMULA_DEPENDENCY_DEPTH = 1000
UNKNOWN = '____UNKNOWN_____'



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 59.0.1-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from __future__ import absolute_import

import sys
import os
import re

# python 2 and python 3 compatibility library
from six import iteritems

from ..configuration import Configuration
from ..api_client import ApiClient
from ..models import *

class UsageApi(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    Ref: https://github.com/swagger-api/swagger-codegen
    """

    def __init__(self, api_client=None):
        config = Configuration()
        if api_client:
            self.api_client = api_client
        else:
            if not config.api_client:
                config.api_client = ApiClient()
            self.api_client = config.api_client

    def get_types(self, **kwargs):
        """
        Get the list of available types that can be used as filter values for the reporting endpoint
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_types(callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :return: UsageTypesV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UsageTypesV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_types_with_http_info(**kwargs)
        else:
            (data) = self.get_types_with_http_info(**kwargs)
            return data

    def get_types_with_http_info(self, **kwargs):
        """
        Get the list of available types that can be used as filter values for the reporting endpoint
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_types_with_http_info(callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :return: UsageTypesV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UsageTypesV1
        """

        all_params = []
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')
        all_params.append('_response_type')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_types" % key
                )
            params[key] = val
        del params['kwargs']

        collection_formats = {}

        path_params = {}

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/usage/types', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type=params.get('_response_type', 'UsageTypesV1'),
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_usage(self, **kwargs):
        """
        Get usage records for the specified time range. Results will be sorted by date if aggregating by day or month, otherwise by amount descending and limited to the first 500 records
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_usage(start_time=start_time_value, end_time=end_time_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str start_time: Start time for the time frame (YYYY-MM-DDThh:mm:ss±hh:mm) (required)
        :param str end_time: End time for the time frame (YYYY-MM-DDThh:mm:ss±hh:mm) (required)
        :param list[str] type: Filter by type of usage (e.g. Analysis, Screenshot)
        :param str source: Filter by any part of the origin label or origin url fields
        :param list[str] identity_id: Filter by the UUID of the user who initiated the request
        :param list[str] aggregate_by: Aggregate the usage records by one or more fields. Cannot aggregate by both Day and Month at the same time.
        :return: UsageOutputListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UsageOutputListV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_usage_with_http_info(**kwargs)
        else:
            (data) = self.get_usage_with_http_info(**kwargs)
            return data

    def get_usage_with_http_info(self, **kwargs):
        """
        Get usage records for the specified time range. Results will be sorted by date if aggregating by day or month, otherwise by amount descending and limited to the first 500 records
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_usage_with_http_info(start_time=start_time_value, end_time=end_time_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str start_time: Start time for the time frame (YYYY-MM-DDThh:mm:ss±hh:mm) (required)
        :param str end_time: End time for the time frame (YYYY-MM-DDThh:mm:ss±hh:mm) (required)
        :param list[str] type: Filter by type of usage (e.g. Analysis, Screenshot)
        :param str source: Filter by any part of the origin label or origin url fields
        :param list[str] identity_id: Filter by the UUID of the user who initiated the request
        :param list[str] aggregate_by: Aggregate the usage records by one or more fields. Cannot aggregate by both Day and Month at the same time.
        :return: UsageOutputListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UsageOutputListV1
        """

        all_params = ['start_time', 'end_time', 'type', 'source', 'identity_id', 'aggregate_by']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')
        all_params.append('_response_type')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_usage" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'start_time' is set
        if ('start_time' not in params) or (params['start_time'] is None):
            raise ValueError("Missing the required parameter `start_time` when calling `get_usage`")
        # verify the required parameter 'end_time' is set
        if ('end_time' not in params) or (params['end_time'] is None):
            raise ValueError("Missing the required parameter `end_time` when calling `get_usage`")


        collection_formats = {}

        path_params = {}

        query_params = []
        if 'start_time' in params:
            query_params.append(('startTime', params['start_time']))
        if 'end_time' in params:
            query_params.append(('endTime', params['end_time']))
        if 'type' in params:
            query_params.append(('type', params['type']))
            collection_formats['type'] = 'multi'
        if 'source' in params:
            query_params.append(('source', params['source']))
        if 'identity_id' in params:
            query_params.append(('identityId', params['identity_id']))
            collection_formats['identityId'] = 'multi'
        if 'aggregate_by' in params:
            query_params.append(('aggregateBy', params['aggregate_by']))
            collection_formats['aggregateBy'] = 'multi'

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/usage', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type=params.get('_response_type', 'UsageOutputListV1'),
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 59.0.1-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from pprint import pformat
from six import iteritems
import re


class ServerSpecOutputV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'component_name': 'str',
        'error': 'bool',
        'message': 'str',
        'system_spec_value': 'str',
        'warning': 'bool'
    }

    attribute_map = {
        'component_name': 'componentName',
        'error': 'error',
        'message': 'message',
        'system_spec_value': 'systemSpecValue',
        'warning': 'warning'
    }

    def __init__(self, component_name=None, error=False, message=None, system_spec_value=None, warning=False):
        """
        ServerSpecOutputV1 - a model defined in Swagger
        """

        self._component_name = None
        self._error = None
        self._message = None
        self._system_spec_value = None
        self._warning = None

        if component_name is not None:
          self.component_name = component_name
        if error is not None:
          self.error = error
        if message is not None:
          self.message = message
        if system_spec_value is not None:
          self.system_spec_value = system_spec_value
        if warning is not None:
          self.warning = warning

    @property
    def component_name(self):
        """
        Gets the component_name of this ServerSpecOutputV1.
        The name of the component. E.G. CPU, Memory...

        :return: The component_name of this ServerSpecOutputV1.
        :rtype: str
        """
        return self._component_name

    @component_name.setter
    def component_name(self, component_name):
        """
        Sets the component_name of this ServerSpecOutputV1.
        The name of the component. E.G. CPU, Memory...

        :param component_name: The component_name of this ServerSpecOutputV1.
        :type: str
        """

        self._component_name = component_name

    @property
    def error(self):
        """
        Gets the error of this ServerSpecOutputV1.
        True if this value is far lower than recommended

        :return: The error of this ServerSpecOutputV1.
        :rtype: bool
        """
        return self._error

    @error.setter
    def error(self, error):
        """
        Sets the error of this ServerSpecOutputV1.
        True if this value is far lower than recommended

        :param error: The error of this ServerSpecOutputV1.
        :type: bool
        """

        self._error = error

    @property
    def message(self):
        """
        Gets the message of this ServerSpecOutputV1.
        An optional message about the spec

        :return: The message of this ServerSpecOutputV1.
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """
        Sets the message of this ServerSpecOutputV1.
        An optional message about the spec

        :param message: The message of this ServerSpecOutputV1.
        :type: str
        """

        self._message = message

    @property
    def system_spec_value(self):
        """
        Gets the system_spec_value of this ServerSpecOutputV1.
        The value of the component

        :return: The system_spec_value of this ServerSpecOutputV1.
        :rtype: str
        """
        return self._system_spec_value

    @system_spec_value.setter
    def system_spec_value(self, system_spec_value):
        """
        Sets the system_spec_value of this ServerSpecOutputV1.
        The value of the component

        :param system_spec_value: The system_spec_value of this ServerSpecOutputV1.
        :type: str
        """

        self._system_spec_value = system_spec_value

    @property
    def warning(self):
        """
        Gets the warning of this ServerSpecOutputV1.
        True if this value is lower than recommended, but not a critical failure

        :return: The warning of this ServerSpecOutputV1.
        :rtype: bool
        """
        return self._warning

    @warning.setter
    def warning(self, warning):
        """
        Sets the warning of this ServerSpecOutputV1.
        True if this value is lower than recommended, but not a critical failure

        :param warning: The warning of this ServerSpecOutputV1.
        :type: bool
        """

        self._warning = warning

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, ServerSpecOutputV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 59.0.1-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from pprint import pformat
from six import iteritems
import re


class FormulaPackageImportInputV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'docs': 'list[FormulaDocInputV1]',
        'formula_package': 'FormulaPackageInputV1',
        'functions': 'list[FunctionInputV1]',
        'sync_token': 'str'
    }

    attribute_map = {
        'docs': 'docs',
        'formula_package': 'formulaPackage',
        'functions': 'functions',
        'sync_token': 'syncToken'
    }

    def __init__(self, docs=None, formula_package=None, functions=None, sync_token=None):
        """
        FormulaPackageImportInputV1 - a model defined in Swagger
        """

        self._docs = None
        self._formula_package = None
        self._functions = None
        self._sync_token = None

        if docs is not None:
          self.docs = docs
        if formula_package is not None:
          self.formula_package = formula_package
        if functions is not None:
          self.functions = functions
        if sync_token is not None:
          self.sync_token = sync_token

    @property
    def docs(self):
        """
        Gets the docs of this FormulaPackageImportInputV1.
        The list of docs to be added or modified

        :return: The docs of this FormulaPackageImportInputV1.
        :rtype: list[FormulaDocInputV1]
        """
        return self._docs

    @docs.setter
    def docs(self, docs):
        """
        Sets the docs of this FormulaPackageImportInputV1.
        The list of docs to be added or modified

        :param docs: The docs of this FormulaPackageImportInputV1.
        :type: list[FormulaDocInputV1]
        """
        if docs is None:
            raise ValueError("Invalid value for `docs`, must not be `None`")

        self._docs = docs

    @property
    def formula_package(self):
        """
        Gets the formula_package of this FormulaPackageImportInputV1.

        :return: The formula_package of this FormulaPackageImportInputV1.
        :rtype: FormulaPackageInputV1
        """
        return self._formula_package

    @formula_package.setter
    def formula_package(self, formula_package):
        """
        Sets the formula_package of this FormulaPackageImportInputV1.

        :param formula_package: The formula_package of this FormulaPackageImportInputV1.
        :type: FormulaPackageInputV1
        """
        if formula_package is None:
            raise ValueError("Invalid value for `formula_package`, must not be `None`")

        self._formula_package = formula_package

    @property
    def functions(self):
        """
        Gets the functions of this FormulaPackageImportInputV1.
        The list of functions to be added or modified

        :return: The functions of this FormulaPackageImportInputV1.
        :rtype: list[FunctionInputV1]
        """
        return self._functions

    @functions.setter
    def functions(self, functions):
        """
        Sets the functions of this FormulaPackageImportInputV1.
        The list of functions to be added or modified

        :param functions: The functions of this FormulaPackageImportInputV1.
        :type: list[FunctionInputV1]
        """
        if functions is None:
            raise ValueError("Invalid value for `functions`, must not be `None`")

        self._functions = functions

    @property
    def sync_token(self):
        """
        Gets the sync_token of this FormulaPackageImportInputV1.
        An arbitrary token (often a date or random ID) that is used during metadata syncs. At the end of a full sync, items with mismatching sync tokens are identified as stale and may be archived using the Datasources clean-up API. This token will be applied to the package, function, and doc items in this import overriding any token value that may have been set in those.

        :return: The sync_token of this FormulaPackageImportInputV1.
        :rtype: str
        """
        return self._sync_token

    @sync_token.setter
    def sync_token(self, sync_token):
        """
        Sets the sync_token of this FormulaPackageImportInputV1.
        An arbitrary token (often a date or random ID) that is used during metadata syncs. At the end of a full sync, items with mismatching sync tokens are identified as stale and may be archived using the Datasources clean-up API. This token will be applied to the package, function, and doc items in this import overriding any token value that may have been set in those.

        :param sync_token: The sync_token of this FormulaPackageImportInputV1.
        :type: str
        """

        self._sync_token = sync_token

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, FormulaPackageImportInputV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



from __future__ import annotations

from collections import defaultdict
from typing import Tuple, List

import pandas as pd
from seeq.spy import _common, _metadata, _status
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy.assets._trees import _constants, _match, _path, _utils
from seeq.spy.assets._trees._pandas import KeyedDataFrame


def validate_and_filter(
        session: Session,
        df_to_validate: pd.DataFrame,
        status: _status.Status,
        stage: str,
        temporal_description='',
        raise_if_all_filtered=False,
        subtract_errors_from_status=False
) -> KeyedDataFrame:
    """
    This is the main validation function. It takes a dataframe as input, and returns a dataframe containing the rows
    for which no errors were found, and a dataframe containing the rows for which errors were found.

    :param session: Login session to use for validation
    :param df_to_validate: DataFrame input
    :param status: Status object to raise exception to if error is found and status.errors='raise'

    :param stage: {'input', 'final'} This must be a key in property_validations. Indicates which
    property validations we want to perform on df_to_validate, depending on whether it is a user-input
    dataframe, a dataframe about to be saved to Tree._dataframe, or some intermediary dataframe. Additionally,
    we only check the tree structure of the dataframe if stage='final'

    :param temporal_description: An adverbial phrase describing the time at which the error(s) were encountered
    :return filtered_df: A validated dataframe with invalid rows removed
    """
    df_to_validate = KeyedDataFrame.of(df_to_validate)
    raise_if_all_filtered = raise_if_all_filtered or stage == 'final'
    error_summaries, error_series = validate(session, df_to_validate, status, stage)
    if len(error_summaries) != 0:
        if status.errors == 'raise':
            df_to_validate['Error Message'] = error_series
            status.df = df_to_validate[error_series != ''].dropna(axis=1, how='all')
            raise_error_summaries(temporal_description, error_summaries, status)
        else:
            keep_items = error_series == ''
            bad_results = df_to_validate[~keep_items].copy()
            bad_results['Error Message'] = error_series[~keep_items]
            filtered_df = df_to_validate[keep_items].reset_index(drop=True)
            _utils.increment_status_df(status, error_items=bad_results, subtract_errors=subtract_errors_from_status)

            if raise_if_all_filtered and filtered_df.empty:
                status.df = bad_results.dropna(axis=1, how='all')
                status.warn('All rows encountered errors and tree could not be constructed')
                raise_error_summaries(temporal_description, error_summaries, status)

            if stage == 'final':
                # We validate again to ensure that self._dataframe will stay valid. Something is fatally wrong
                # with validation if the following code is reached.
                further_error_summaries, error_series = validate(session, filtered_df)
                if len(further_error_summaries) != 0:
                    filtered_df['Error Message'] = error_series
                    status.df = filtered_df[error_series != ''].dropna(axis=1, how='all')
                    raise_error_summaries('while validating tree', error_summaries, status)

            warn_error_summaries(temporal_description, error_summaries, status)

            return rectify_column_order(filtered_df)
    else:
        return rectify_column_order(df_to_validate)


def validate(
        session: Session, df: KeyedDataFrame, status: _status.Status = None, stage='final'
) -> Tuple[List[str], pd.Series]:
    if status is None:
        status = _status.Status(quiet=True)

    error_summaries_properties, error_series_properties = validate_properties(session, df, stage)
    if stage == 'final':
        # Only do tree validation in the final validation step, i.e., when this df represents a tree
        # Don't do tree validation on rows that had property errors
        ignore_rows = error_series_properties != ''
        error_summaries_tree, error_series_tree = validate_tree_structure(df, ignore_rows=ignore_rows)

        ignore_rows = ignore_rows | (error_series_tree != '')
        error_summaries_dependencies, error_series_dependencies = validate_dependencies(df, status,
                                                                                        ignore_rows=ignore_rows)
    else:
        error_summaries_tree, error_series_tree = list(), pd.Series('', index=df.index)
        error_summaries_dependencies, error_series_dependencies = list(), pd.Series('', index=df.index)

    error_summaries = error_summaries_properties + error_summaries_tree + error_summaries_dependencies
    error_series = update_error_msg(error_series_properties, error_series_tree)
    error_series = update_error_msg(error_series, error_series_dependencies)

    return error_summaries, error_series


def get_error_message(temporal_description, error_summaries):
    count = len(error_summaries)
    msg = '%s encountered %s' % ('1 error was' if count == 1 else f'{count} errors were', temporal_description)
    msg += ':\n- ' + '\n- '.join(error_summaries[:_constants.MAX_ERRORS_DISPLAYED])
    if count > _constants.MAX_ERRORS_DISPLAYED:
        additional_errors = count - _constants.MAX_ERRORS_DISPLAYED
        msg += f'\n- {additional_errors} additional issue{"s" if additional_errors > 1 else ""} found.'
    return msg


def raise_error_summaries(temporal_description, error_summaries, status):
    msg = get_error_message(temporal_description, error_summaries)
    status.exception(SPyRuntimeError(msg), throw=True, use_error_message=True)


def warn_error_summaries(temporal_description, error_summaries, status):
    msg = get_error_message(temporal_description, error_summaries)
    status.warn(msg)
    status.update()


def update_error_msg(old_msg, new_msg):
    if new_msg is None or isinstance(new_msg, str) and new_msg == '':
        return old_msg
    out = old_msg + ' ' + new_msg
    if isinstance(out, pd.Series):
        return out.str.strip()
    else:
        return out.strip()


def validate_tree_structure(df, ignore_rows=None):
    # Asserts that:
    # - The tree is non-empty
    # - The root doesn't have a path, and is the only item with depth 1
    # - The dataframe is sorted by path
    # - There are no missing assets referenced in paths
    # - Paths reflect names of preceding items
    # - Depths reflects lengths of paths

    size = len(df)
    if size == 0:
        return ['Tree must be non-empty.'], pd.Series(dtype='string')

    error_series = pd.Series('', index=df.index)
    error_summaries = []
    if ignore_rows is None:
        ignore_rows = pd.Series(False, index=df.index)

    prev_path = list()
    prev_type = 'Asset'
    _path.decorate_with_full_path(df)
    for i, row in df.iterrows():
        if error_series.iloc[i]:
            # Node has an error message already due to a bad ancestor
            continue

        depth = row.Depth
        this_path = row['Full Path List']

        try:
            if ignore_rows[i]:
                # Ignore tree errors on this row because of a property validation error
                # We still want invalidate its children if possible, so we raise an assertion error with no message
                assert False, ''

            assert depth == len(this_path), 'Item\'s depth does not match its path.'

            if i == 0:
                assert len(this_path) == 1, 'The root of the tree cannot be assigned a path.'
                # The following assertion will be handled differently to include node names in the error message
                assert (df['Full Path List'].iloc[1:].apply(len) != 1).all(), 'A tree can only have one root but ' \
                                                                              'multiple were given: '
            else:
                assert depth >= 1, 'Only depths greater or equal to 1 are valid.'

            if depth <= len(prev_path):
                assert prev_path[:depth - 1] == this_path[:depth - 1], 'Item\'s position in tree ' \
                                                                       'does not match its path.'
                assert prev_path[depth - 1] != this_path[depth - 1], 'Item has the same name and path ' \
                                                                     'of another item in the tree.'
                assert prev_path[depth - 1] < this_path[depth - 1], 'Item is not stored in proper ' \
                                                                    'position sorted by path.'
            else:
                assert depth == len(prev_path) + 1, 'Item has an ancestor not stored in this tree.'
                assert prev_path[:depth - 1] == this_path[:depth - 1], 'Item\'s position in tree ' \
                                                                       'does not match its path.'
                assert prev_type == 'Asset', 'Item\'s parent must be an Asset.'

            prev_path = this_path
            prev_type = row.Type

        except AssertionError as e:
            message = str(e)
            if message.startswith('A tree can only have one root'):
                roots = df.Depth == 1
                message += '"%s".' % '\", \"'.join(df.Name[roots])
                error_series[roots] = message
                error_series[~roots] = 'Item\'s parent is invalid.'
                error_summaries.append(message)
                break
            error_series[i] = message
            children = df['Full Path List'].apply(
                lambda path: len(path) > len(this_path) and path[:len(this_path)] == this_path)
            error_series[children] = 'Item\'s parent is invalid.'
            if message:
                error_summaries.append(f'Invalid item with path "{" >> ".join(this_path)}": ' + message)

    _path.remove_full_path(df)

    return error_summaries, error_series


def validate_dependencies(df, status: _status.Status, ignore_rows=None):
    error_series = pd.Series('', index=df.index)
    error_summaries = set()
    if ignore_rows is None:
        ignore_rows = pd.Series(False, index=df.index)

    if len(df) == 0 or ignore_rows.all():
        return list(), pd.Series('', index=df.index)

    if ignore_rows.any():
        df = df.loc[~ignore_rows]
    dependency_graph = defaultdict(set)

    for node in _match.TreeNode.of(df):  # type: _match.TreeNode
        dependencies = list()
        if 'Formula Parameters' in df.columns and pd.api.types.is_dict_like(df.loc[node.index, 'Formula Parameters']):
            dependencies.extend(df.loc[node.index, 'Formula Parameters'].values())
        if 'Measured Item' in df.columns and isinstance(df.loc[node.index, 'Measured Item'], str):
            dependencies.append(df.loc[node.index, 'Measured Item'])
        if 'Bounding Condition' in df.columns and isinstance(df.loc[node.index, 'Bounding Condition'], str):
            dependencies.append(df.loc[node.index, 'Bounding Condition'])
        for dependency in dependencies:
            if not isinstance(dependency, str) or _common.is_guid(dependency):
                continue
            try:
                resolved_dependency = node.resolve_reference(dependency)
                dependency_graph[node].add(resolved_dependency)
            except SPyRuntimeError as e:
                message = e.args[0]
                error_series[node.index] = update_error_msg(error_series[node.index], message)
                error_summaries.add(message)
        if 'Roll Up Parameters' in df.columns and isinstance(df.loc[node.index, 'Roll Up Parameters'], str):
            try:
                resolved_dependencies = node.resolve_references(
                    df.loc[node.index, 'Roll Up Parameters'])
                dependency_graph[node].update(resolved_dependencies)
                if len(resolved_dependencies) == 0:
                    status.warn(f'Roll up parameter "{df.loc[node.index, "Roll Up Parameters"]}" does not match any '
                                f'items in the tree.')
            except SPyRuntimeError as e:
                message = e.args[0]
                error_series[node.index] = update_error_msg(error_series[node.index], message)
                error_summaries.add(message)

    node_results = dict()

    def check_for_invalid_dependencies(node, depth=0):
        if depth > _constants.MAX_FORMULA_DEPENDENCY_DEPTH:
            raise SPyRuntimeError('Circular dependencies detected. Check that the formula parameters for the specified '
                                  'items do not create a loop.')

        if error_series[node.index] != '' or ignore_rows[node.index]:
            node_results[node] = False
        if node in node_results:
            return node_results[node]
        for dependency in dependency_graph[node]:
            if not check_for_invalid_dependencies(dependency, depth + 1):
                error_series[node.index] = 'Item references an invalid tree item.'
                break
        result = error_series[node.index] == ''
        node_results[node] = result
        return result

    for node in list(dependency_graph.keys()):
        check_for_invalid_dependencies(node)

    return list(error_summaries), error_series


def validate_properties(session: Session, df, stage):
    """
    :param session: Login session to use for validation
    :param df: The dataframe to be validated for errors related to presence of properties, type of properties,
    and ill-defined properties
    """
    error_series = pd.Series('', index=df.index)
    error_message_map = dict()  # maps error messages to the rows that encountered the error
    for index, node in df.iterrows():
        errors = validate_node_properties(session, node, stage)
        for error in errors:
            _common.get(error_message_map, error, default=list(), assign_default=True).append((index, node))
        if errors:
            error_series[index] = ' '.join(errors)

    error_summaries = collect_error_messages(error_message_map)

    return error_summaries, error_series


def collect_error_messages(error_message_map):
    def _get_row_description(index, row):
        description_properties = dict()
        # Prefer Friendly Name over Name
        if _common.present(row, 'Friendly Name'):
            description_properties['friendly name'] = row['Friendly Name']
        elif _common.present(row, 'Name'):
            description_properties['name'] = row['Name']
        # If a Name or Friendly Name has been found, add a Path too if it is present
        if len(description_properties) != 0 and (_common.present(row, 'Path') or _common.present(row, 'Asset')):
            description_properties['path'] = _path.determine_path(row)
        # Use ID next if it is present
        if len(description_properties) == 0 and _common.present(row, 'ID'):
            description_properties['ID'] = row['ID']

        # Use index if none of the above are present
        if len(description_properties) == 0:
            return f'The item with index {index}'
        else:
            return 'The item with ' + ' and '.join([f'{prop_name} "{prop_value}"' for prop_name, prop_value in
                                                    description_properties.items()])

    def _get_row_descriptiveness_score(row):
        _, row = row
        if _common.present(row, 'Name') or _common.present(row, 'Friendly Name'):
            if _common.present(row, 'Path') or _common.present(row, 'Asset'):
                return 3
            else:
                return 2
        elif _common.present(row, 'ID'):
            return 1
        else:
            return 0

    def _get_most_descriptive_row(_rows):
        index, row = max(_rows, key=_get_row_descriptiveness_score)
        return _get_row_description(index, row)

    collected_messages = list()
    for message, rows in error_message_map.items():
        if len(collected_messages) >= _constants.MAX_ERRORS_DISPLAYED:
            # No need to fuss with error messaging formatting that won't be displayed. We pass in placeholder string
            collected_messages.extend(('' for _ in range(len(error_message_map) - _constants.MAX_ERRORS_DISPLAYED)))
            break
        if len(rows) == 1:
            collected_messages.append(f'{_get_row_description(*rows[0])} has the following issue: {message}')
        else:
            collected_messages.append(f'{_get_most_descriptive_row(rows)} and {len(rows) - 1} other '
                                      f'item{"s" if len(rows) > 2 else ""} has the following issue: {message}')
    return collected_messages


def validate_node_properties(session, node, stage):
    def has_bad_type(column, dtype):
        if _common.present(node, column):
            datum = _utils.safe_int_cast(node[column])
            try:
                _common.validate_argument_types([(datum, '', dtype)])
            except TypeError:
                return True
        return False

    def dtype_names(dtype):
        if isinstance(dtype, tuple):
            return tuple(x.__name__ for x in dtype)
        return dtype.__name__

    errors = [f"The property '{column}' must have one of the following types: {dtype_names(dtype)}."
              for column, dtype in _constants.dataframe_dtypes.items() if has_bad_type(column, dtype)]

    # The conditions in property_validations assume that values have the correct datatype
    # Therefore, return only type errors if they exist.
    if errors:
        return errors
    errors += [message for requirement, message in property_validations(session, node, stage) if not requirement]

    return errors


def no_repeated_nested_paths(path_list, name, recurse=True):
    if len(path_list) == 0:
        return True
    if path_list[-1] == name:
        return False
    return no_repeated_nested_paths(path_list[:-1], path_list[-1]) if recurse else True


def rectify_column_order(df):
    standard_columns = [col for col in _constants.dataframe_columns if col in df.columns]
    extra_columns = sorted([col for col in df.columns if col not in _constants.dataframe_columns])
    columns = standard_columns + extra_columns
    return df[columns]


def property_validations(session: Session, node, stage):
    if stage == 'input':
        return [
            (_common.get(node, 'ID') or _common.get(node, 'Name') or _common.get(node, 'Friendly Name'),
             "The property 'Name' or 'Friendly Name' is required for all nodes without ID."),
            (not _common.get(node, 'Type') or _common.get(node, 'Formula') or _common.get(node, 'Roll Up Statistic') or
             ('Condition' not in node['Type'] and 'Signal' not in node['Type']) or _common.get(node, 'ID'),
             "Stored Signals and Conditions are not yet supported. "
             "All Signals and Conditions require either a formula or an ID."),
            (not _common.present(node, 'Formula Parameters') or _common.present(node, 'Formula'),
             "Must have a Formula if Formula Parameters are defined."),
            (not (_common.present(node, 'ID') or _common.present(node, 'Referenced ID')) or session.client,
             "Must log in via spy.login() before inserting an item via ID or Referenced ID."),
            (not _common.present(node, 'ID') or _common.is_guid(node['ID']),
             f"The property 'ID' must be a valid GUID. Given: '{_common.get(node, 'ID')}'"),
            (not _common.present(node, 'Referenced ID') or _common.is_guid(node['Referenced ID']),
             f"The property 'Referenced ID' must be a valid GUID. Given: '{_common.get(node, 'Referenced ID')}'"),
            (not (_common.get(node, 'Path') and _common.get(node, 'Name')) or
             no_repeated_nested_paths(_common.path_string_to_list(node['Path']), node['Name']),
             "Paths with repeated names are not valid."),
            ((not (_common.present(node, 'Formula') or _common.present(node, 'Formula Parameters'))
              or not _common.get(node, 'Type') == 'Asset'),
             "Assets cannot have formulas or formula parameters."),
            (not _common.get(node, 'Type') == 'Asset' or not _common.present(node, 'Roll Up Statistic'),
             "Assets cannot be roll ups."),
            (_is_node_metric_valid(stage, node, session)),
            (_common.present(node, 'Roll Up Statistic') == _common.present(node, 'Roll Up Parameters'),
             "Roll ups must specify both 'Roll Up Statistic' and 'Roll Up Parameters'."),
            (not _common.present(node, 'Roll Up Statistic') or not _common.present(node, 'Formula'),
             "Roll ups cannot have a formula."),
            (not _common.present(node, 'Roll Up Statistic')
             or node['Roll Up Statistic'].casefold() in (fn.statistic for fn in _common.ROLL_UP_FUNCTIONS),
             f"Roll up statistic '{_common.get(node, 'Roll Up Statistic')}' not found. Valid options: "
             f"{', '.join(set(fn.statistic.capitalize() for fn in _common.ROLL_UP_FUNCTIONS))}")
        ]
    elif stage == 'final':
        return [
            (_common.get(node, 'Formula') or _common.get(node, 'Type') or _common.get(node, 'Roll Up Statistic'),
             "The property 'Type' is required for all items without formulas or roll-up statistics."),
            (not _common.get(node, 'Type') or _common.get(node, 'Formula') or _common.get(node, 'Roll Up Statistic') or
             ('Condition' not in node['Type'] and 'Signal' not in node['Type']),
             "Stored Signals and Conditions are not yet supported. All Signals and Conditions require a formula."),
            (not _common.present(node, 'Formula Parameters') or _common.present(node, 'Formula'),
             "Must have a Formula if Formula Parameters are defined."),
            (_common.present(node, 'Name'),
             "The property 'Name' is required."),
            (_common.present(node, 'Path'),
             "The property 'Path' is required."),
            (_common.present(node, 'Depth'),
             "The property 'Depth' is required."),
            (not _common.present(node, 'Name') or not _common.present(node, 'Path') or
             no_repeated_nested_paths(_common.path_string_to_list(node['Path']), node['Name'], recurse=False),
             "Paths with repeated names are not valid."),
            ((not (_common.present(node, 'Formula') or _common.present(node, 'Formula Parameters'))
              or not _common.get(node, 'Type') == 'Asset'),
             "Assets cannot have formulas or formula parameters."),
            (not _common.get(node, 'Type') == 'Asset' or not _common.present(node, 'Roll Up Statistic'),
             "Assets cannot be roll ups."),
            (_is_node_metric_valid(stage, node, session)),
            (_common.present(node, 'Roll Up Statistic') == _common.present(node, 'Roll Up Parameters'),
             "Roll ups must specify both 'Roll Up Statistic' and 'Roll Up Parameters'."),
            (not _common.present(node, 'Roll Up Statistic') or not _common.present(node, 'Formula'),
             "Roll ups cannot have a formula."),
        ]


def _is_node_metric_valid(stage: str, node: pd.Series, session: Session) -> (bool, str):
    if not _common.present(node, 'Type') or 'Metric' not in _common.get(node, 'Type'):
        return True, None  # Okay because this isn't a metric
    if not _common.present(node, 'Measured Item'):
        if stage == 'input' and not _common.present(node, 'ID'):
            return False, 'Metrics must have a Measured Item or an ID.'
        elif stage == 'final':
            return False, 'Metrics must have a Measured Item.'
    if _common.present(node, 'Metric Neutral Color'):
        neutral_color = _common.get(node, 'Metric Neutral Color')
        if not neutral_color.startswith('#') or not (len(neutral_color) == 4 or len(neutral_color) == 7):
            return False, "Metric neutral color must start with a '#' character and be a valid hex value."
    if _common.present(node, 'Statistic'):
        try:
            _common.statistic_to_aggregation_function(node['Statistic'])
        except SPyValueError as e:
            return False, str(e)
    if _common.present(node, 'Thresholds'):
        thresholds = _common.get(node, 'Thresholds')
        if isinstance(thresholds, dict):
            session_validated = Session.validate(session)
            if session_validated:
                try:
                    _metadata.convert_threshold_levels_from_system(session_validated, thresholds, node['Name'])
                except SPyRuntimeError as e:
                    return False, str(e)
        # Other valid `thresholds` types include a List, but that's likely directly from Appserver so do not
        # validate it here. All other types will have been caught by the dataframe_dtypes check.
    return True, None



import argparse
import datetime
import itertools
from datetime import date
from hashlib import shake_128
from typing import List, Union

import pandas as pd

from seeq import spy
from seeq.base.seeq_names import SeeqNames
from seeq.spy.workbooks import Analysis, AnalysisWorksheet

AGILE_FILTER_5_MIN = '.agileFilter(5min)'
RUNNING_DELTA = '.runningDelta().max(0)'
ROOT_NODE = 'Performance Monitors'
DATASOURCES_WORKBOOK_SUFFIX = ' - Datasources'
DATASOURCE = 'Seeq Monitoring Script'
SEEQ_PERSONNEL_USER_GROUP_GUID = '792CFDF4-94B3-44B3-A2CA-7157230CC1E2'
CUSTOMER_MONITORS_FOLDER_ID = '54BAFDB1-E742-42C0-B912-8902281E722F'
# These are the columns that are returned by spy.push in the result dataframe when pushing signal metadata
PUSH_RESULT_COLUMNS = ['Datasource Class', 'Datasource ID', 'Data ID', 'ID', 'Push Result']


def monitor_group_definitions(server: str):
    return {
        'System Overview': [
            {'path': ['Cpu', 'System', 'UsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', 'Appserver', 'ServerLoadPercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'Jobs', 'User', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Users', 'ActiveAuthenticatedUsers', 'Last15Min'],
             'name': 'Gauge'},
            {'path': ['Processes', 'Appserver', 'Uptime'],
             'name': 'Gauge'},
        ],
        'Cpu Overview': [
            {'path': ['Cpu', 'System', 'UsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Cpu', 'Appserver', 'UsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'Appserver', 'CpuUsagePercent'],
             'name': 'Gauge'},
        ],
        'Cpu Usage': [
            {'path': ['Processes', server, 'Appserver', 'CpuUsagePercent'],
             'name': 'Gauge'},
            {'path': ['Processes', server, 'JvmLink', 'CpuUsagePercent'],
             'name': 'Gauge'},
            {'path': ['Processes', server, 'Nginx', 'CpuUsagePercent'],
             'name': 'Gauge'},
            {'path': ['Processes', server, 'Postgres', 'CpuUsagePercent'],
             'name': 'Gauge'},
            {'path': ['Processes', server, 'Renderer', 'CpuUsagePercent'],
             'name': 'Gauge'},
            {'path': ['Processes', server, 'ScalarCache', 'CpuUsagePercent'],
             'name': 'Gauge'},
            {'path': ['Processes', server, 'ScreenshotWorkers', 'CpuUsagePercent'],
             'name': 'Gauge'},
            {'path': ['Processes', server, 'SeriesCache', 'CpuUsagePercent'],
             'name': 'Gauge'},
            {'path': ['Processes', server, 'SeriesCachePostgres', 'CpuUsagePercent'],
             'name': 'Gauge'},
            {'path': ['Processes', server, 'SupervisorCore', 'CpuUsagePercent'],
             'name': 'Gauge'},
        ],
        'Garbage Collector': [
            {'path': ['Memory', 'GarbageCollector', 'Pause', 'Timer'],
             'name': 'Count',
             'calculated_signals': ['agile_filter']},
            {'path': ['Memory', 'GarbageCollector', 'Pause', 'Timer'],
             'name': 'Mean',
             'calculated_signals': ['agile_filter']},
            {'path': ['Memory', 'GarbageCollector', 'Pause', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']},
            {'path': ['Memory', 'GarbageCollector', 'TotalPauseTime'],
             'name': 'Meter',
             'calculated_signals': ['agile_filter']},
        ],
        'Memory overview': [
            {'path': ['Processes', server, 'Appserver', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Memory', 'Appserver', 'AvailableMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Memory', 'System', 'AvailablePhysicalMemoryMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Cache', 'NonSeries', 'Memory'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
        ],
        'Memory usage': [
            {'path': ['Processes', server, 'Appserver', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'JvmLink', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'Nginx', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'Postgres', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'Renderer', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'ScalarCache', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'ScreenshotWorkers', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'SeriesCache', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'SeriesCachePostgres', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'SupervisorCore', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
        ],
        'Disk Space': [
            {'path': ['HardDisk', 'Usage', 'AvailableBytes', 'DataFolder'],
             'name': 'Gauge'},
        ],
        'Disk MyData': [
            {'path': ['HardDisk', 'Usage', 'Folders', 'MyData', 'FileCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'MyData', 'FolderCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'MyData', 'Size'],
             'name': 'Gauge'},
        ],
        'Disk Cache': [
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Content', 'FileCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Series', 'FileCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Thumbnail', 'FileCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Content', 'FolderCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Series', 'FolderCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Thumbnail', 'FolderCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Content', 'Size'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Series', 'Size'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Thumbnail', 'Size'],
             'name': 'Gauge'},
        ],
        'Long running requests': [
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Running'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Completed'],
             'name': 'Meter',
             'calculated_signals': ['running_delta_agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Submitted'],
             'name': 'Meter',
             'calculated_signals': ['running_delta_agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Queued', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Duration', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']},
        ],
        'Users': [
            {'path': ['Users', 'ActiveAuthenticatedUsers', 'Last15Min'],
             'name': 'Gauge'},
            {'path': ['Users', 'BrowserTabs'],
             'name': 'Gauge'},
            {'path': ['Users', 'AuthenticatedUsers'],
             'name': 'Gauge'},
        ],
        'User jobs': [
            {'path': ['Threads', 'Jobs', 'User', 'Running'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'Jobs', 'User', 'Completed'],
             'name': 'Meter',
             'calculated_signals': ['running_delta_agile_filter']},
            {'path': ['Threads', 'Jobs', 'User', 'Submitted'],
             'name': 'Meter',
             'calculated_signals': ['running_delta_agile_filter']},
            {'path': ['Threads', 'Jobs', 'User', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'Jobs', 'User', 'Queued', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'Jobs', 'User', 'Duration', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']},
        ],
        'Thread Pool Queues': [
            {'path': ['Threads', 'HttpServer', 'Jobs', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'Request', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'Default', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'Initialization', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']}
        ],
        'Client Websocket': [
            {'path': ['Threads', 'ClientWebSocket', 'Messages', 'Running'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'AgentWebSocket', 'Messages', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'ClientWebSocket', 'Messages', 'Completed'],
             'name': 'Meter',
             'calculated_signals': ['running_delta_agile_filter']},
            {'path': ['Threads', 'ClientWebSocket', 'Messages', 'Duration', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']}
        ],
        'Network': [
            {'path': ['Network', 'PubSub', 'Channels', 'BroadcastChannel', 'Subscribers'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Network', 'AgentWebSocket', 'ConnectedAgents'],
             'name': 'Gauge'},
            {'path': ['Network', 'HttpServer', 'RestAPI', 'Samples'],
             'name': 'Meter',
             'calculated_signals': ['running_delta_agile_filter']},
            {'path': ['Network', 'HttpServer', 'RestAPI', 'Capsules'],
             'name': 'Meter',
             'calculated_signals': ['running_delta_agile_filter']}
        ],
        'Screenshots': [
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Timing', 'TotalRenderingTime', 'Timer'],
             'name': 'Count',
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Timing', 'TotalRenderingTime', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Timing', 'Queued', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Timing', 'DataDownload', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Result', 'Aborted'],
             'name': 'Meter',
             'calculated_signals': ['running_delta_agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Result', 'Successful'],
             'name': 'Meter',
             'calculated_signals': ['running_delta_agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Queued'],
             'name': 'Gauge',
             'calculated_signals': ['running_delta_agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Running'],
             'name': 'Gauge',
             'calculated_signals': ['running_delta_agile_filter']}
        ],
        'Thumbnails': [
            {'path': ['Rendering', 'HeadlessCapture', 'Thumbnail', 'Timing', 'TotalRenderingTime', 'Timer'],
             'name': 'Count',
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Thumbnail', 'Timing', 'TotalRenderingTime', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Thumbnail', 'Timing', 'Queued', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Thumbnail', 'Result', 'Aborted'],
             'name': 'Meter',
             'calculated_signals': ['running_delta_agile_filter']}
        ],
        'Persistent Signal Cache': [
            {'path': ['Cache', 'Persistent', 'Signal', 'Read', 'Sample'],
             'name': 'Meter',
             'calculated_signals': ['running_delta_agile_filter']},
            {'path': ['Cache', 'Persistent', 'Signal', 'Read', 'Time', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']},
            {'path': ['Cache', 'Persistent', 'Signal', 'Write', 'Time', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']}
        ],
        'Persistent Condition Cache': [
            {'path': ['Cache', 'Persistent', 'Condition', 'Read', 'Capsule'],
             'name': 'Meter',
             'calculated_signals': ['running_delta_agile_filter']},
            {'path': ['Cache', 'Persistent', 'Condition', 'Read', 'Time', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']},
            {'path': ['Cache', 'Persistent', 'Condition', 'Write', 'Time', 'Timer'],
             'name': 'P95',
             'calculated_signals': ['agile_filter']}
        ],
        'Server Uptime': [
            {'path': ['Processes', 'Appserver', 'Uptime'],
             'name': 'Gauge'},
        ],
    }


EXCLUDED_DATASOURCES = {
    'Auth',
    'CassandraV2',
    'ConnectedCount',
    'FileFolders',
    'FileSignal',
    'Monitors'
    'Rx',
    'Seeq - FileSignal'
    'Seeq Metadata',
    'SeeqMonitors',
    'Seeq - Signal Datasource',
}

COLOR_CODES = [
    '#0910a4', '#cfd1fc', '#c01311', '#f19695', '#8B6969', '#D2C3C3', '#d94bff', '#eca5ff', '#7e86d6', '#bec2ea',
    '#c70000', '#ff6363', '#0075d2', '#69bdff', '#e6d50a', '#f9f07e', '#ff5959', '#ffacac', '#ffd8ba', '#ffebdc',
    '#ff7d1a', '#ffbe8c', '#e3e751', '#f1f3a8', '#00ced1', '#b4feff', '#0a0dd8', '#7779f9', '#08d2fb', '#b5f1fe',
    '#C67171', '#E7C4C4', '#F4A460', '#FAD8BA', '#8B7355', '#D4C8B9', '#777733', '#D8D8A4', '#DB2645', '#F1A8B5',
    '#5C246E', '#CC97DD', '#23238E', '#9898E6', '#7e86d6', '#bec2ea', '#d94bff', '#eca5ff', '#c70000', '#ff6363',
]
MANY_COLOR_CODES = COLOR_CODES * 13


def assign_color_to_stored_signals(results: pd.DataFrame, lane: int, offset: int = 1) -> None:
    results['Color'] = MANY_COLOR_CODES[2 * lane - offset]


def customize_stored_signals_trend_display(results: pd.DataFrame, lane: int) -> pd.DataFrame:
    results['Line Width'] = 0.5
    results['Lane'] = lane
    results['Line Style'] = 'Solid'
    assign_color_to_stored_signals(results, lane)
    return results


def create_agile_filter_calculated_signal(name: str, formula_guid: str, lane: int) -> pd.DataFrame:
    signal_name = name.replace(' >> ', '.') + AGILE_FILTER_5_MIN + f'-{variable_digest(formula_guid, 3)}'
    return pd.DataFrame([{
        SeeqNames.Properties.name: signal_name,
        'Type': SeeqNames.Types.signal,
        SeeqNames.Properties.formula: f'$s{AGILE_FILTER_5_MIN}',
        'Formula Parameters': {'$s': formula_guid},
        'Line Width': 1.5,
        'Lane': lane,
        'Color': MANY_COLOR_CODES[2 * lane - 2],
    }])


def create_running_delta_agile_filter_calculated_signal(name: str, formula_guid: str, lane: int) -> pd.DataFrame:
    signal_name = name.replace(' >> ', '.') + \
                  RUNNING_DELTA + AGILE_FILTER_5_MIN + f'-{variable_digest(formula_guid, 3)}'
    return pd.DataFrame([{
        SeeqNames.Properties.name: signal_name,
        'Type': SeeqNames.Types.signal,
        SeeqNames.Properties.formula: f'$s{RUNNING_DELTA}{AGILE_FILTER_5_MIN}',
        'Formula Parameters': {'$s': formula_guid},
        'Line Width': 1.5,
        'Lane': lane,
        'Color': MANY_COLOR_CODES[2 * lane - 2],
    }])


CALCULATED_SIGNAL_BUILDERS = {
    'agile_filter': create_agile_filter_calculated_signal,
    'running_delta_agile_filter': create_running_delta_agile_filter_calculated_signal
}


def push_calculated_signal_metadata(calculated_signals_to_trend: pd.DataFrame, workbook_id: str) -> pd.DataFrame:
    return spy.push(
        metadata=calculated_signals_to_trend.reset_index(drop=True),
        datasource=DATASOURCE,
        workbook=workbook_id,
        quiet=True
    ).set_index('Name', drop=False)


def set_trend_labels(worksheet, label, location) -> None:
    stores = worksheet.current_workstep().get_workstep_stores()
    stores['sqTrendStore'] = {'labelDisplayConfiguration': {label: location}}


def set_auto_update_display_range(worksheet) -> None:
    stores = worksheet.current_workstep().get_workstep_stores()
    stores['sqDurationStore']['autoUpdate']['mode'] = 'MANUAL'
    stores['sqDurationStore']['autoUpdate']['manualInterval'] = {'value': 5, 'units': 'min', 'valid': True}


def set_worksheet_display_items(worksheet: AnalysisWorksheet, signals_to_trend: List) -> None:
    signals_to_trend = pd.concat(signals_to_trend)
    signals_to_trend.reset_index(drop=True, inplace=True)
    worksheet.display_items = signals_to_trend
    set_trend_labels(worksheet, 'name', 'lane')
    set_auto_update_display_range(worksheet)
    print(f'Adding signals to trend')


def create_worksheet_for_monitor(workbook: Analysis, monitor: str, start: datetime.date, end: datetime.date):
    worksheet = workbook.worksheet(monitor)
    worksheet.display_range = {
        'Start': start,
        'End': end
    }
    return worksheet


def get_stored_signal_lane_def(stored_signal_metadata: pd.DataFrame, signal: dict, lane: int) -> pd.DataFrame:
    signal_name = build_name_str(signal['path'], signal['name'])
    results = stored_signal_metadata[stored_signal_metadata['Name'].str.lower() == signal_name.lower()].copy()
    results['Lane'] = lane
    if 'calculated_signals' in signal:
        customize_stored_signals_trend_display(results, lane)
    else:
        assign_color_to_stored_signals(results, lane, 2)
    return results


def get_calculated_signal_lane_def(stored_signal_metadata: pd.DataFrame, signal: dict, lane: int) \
        -> Union[pd.DataFrame, None]:
    if 'calculated_signals' not in signal:
        return None
    calculated_signals = [CALCULATED_SIGNAL_BUILDERS[filter_name] for filter_name in signal['calculated_signals']]
    signal_path = build_path_str(signal['path'])
    return pd.concat([
        filter_func(signal_path, stored_signal, lane)
        for stored_signal in stored_signal_metadata['ID']
        for filter_func in calculated_signals
    ]).set_index('Name', drop=False)


def merge_push_results(lane_defs: List[Union[pd.DataFrame, None]], pushed_metadata: pd.DataFrame) -> List[pd.DataFrame]:
    return [
        lane_def.merge(
            pushed_metadata[PUSH_RESULT_COLUMNS],
            left_index=True,
            right_index=True
        )
        for lane_def in lane_defs
        if lane_def is not None
    ]


def create_regular_monitor_worksheets(workbook: Analysis,
                                      stored_signal_metadata: pd.DataFrame,
                                      two_weeks_ago: datetime.date,
                                      today: datetime.date,
                                      server: str) -> None:
    worksheet_definitions = {}
    for monitor_group, monitor_definitions in monitor_group_definitions(server).items():
        worksheet = create_worksheet_for_monitor(workbook, monitor_group, two_weeks_ago, today)
        stored_signal_lane_defs = [get_stored_signal_lane_def(stored_signal_metadata, monitor_definition, i + 1)
                                   for i, monitor_definition
                                   in enumerate(monitor_definitions)]
        for stored_signal_lane_def, monitor_definition in zip(stored_signal_lane_defs, monitor_definitions):
            if stored_signal_lane_def.empty:
                name = build_name_str(monitor_definition['path'], monitor_definition['name'])
                raise RuntimeError(f"A Stored Signal {name} wasn't found. Arguments to --server and --fqdn probably "
                                   f"need to differ")
            for name in stored_signal_lane_def['Name']:
                print(f'Found Stored Signal {name}')
        calculated_signal_lane_defs = [
            get_calculated_signal_lane_def(stored_signal_lane_def, monitor_definition, i + 1)
            for i, (stored_signal_lane_def, monitor_definition)
            in enumerate(zip(stored_signal_lane_defs, monitor_definitions))
        ]
        for calculated_signal_lane_def in calculated_signal_lane_defs:
            if calculated_signal_lane_def is not None:
                for name in calculated_signal_lane_def['Name']:
                    print(f'Created Calculated Signal {name}')

        worksheet_definitions[monitor_group] = {
            'worksheet': worksheet,
            'stored_signal_lane_defs': stored_signal_lane_defs,
            'calculated_signal_lane_defs': calculated_signal_lane_defs
        }

    # some worksheets monitor the same signals, so we need to drop duplicates to successfully push everything
    calculated_signal_metadata = pd.concat([
        calculated_signal_lane_def
        for worksheet_definition in worksheet_definitions.values()
        for calculated_signal_lane_def in worksheet_definition['calculated_signal_lane_defs']
        if calculated_signal_lane_def is not None
    ]).drop_duplicates(['Name'])
    print(f'Pushing calculated signals for workbook {workbook.id}')
    pushed_calculated_signal_metadata = push_calculated_signal_metadata(
        calculated_signal_metadata, workbook.id
    )
    print(f'Finished pushing calculated signals for workbook {workbook.id}')

    worksheet_definitions = {
        monitor_group: {
            **worksheet_definition,
            **{
                'calculated_signal_lane_defs': merge_push_results(
                    worksheet_definition['calculated_signal_lane_defs'],
                    pushed_calculated_signal_metadata
                )
            }
        }
        for monitor_group, worksheet_definition in worksheet_definitions.items()
    }
    for monitor_group, worksheet_definition in worksheet_definitions.items():
        worksheet = worksheet_definition['worksheet']
        lane_defs = [*worksheet_definition['stored_signal_lane_defs'],
                     *worksheet_definition['calculated_signal_lane_defs']]
        if len(lane_defs) > 0:
            set_worksheet_display_items(worksheet, lane_defs)


def create_datasource_monitor_worksheets(workbook: Analysis,
                                         two_weeks_ago: datetime.date,
                                         today: datetime.date,
                                         customer: str,
                                         fqdn: str) -> None:
    # Datasources are customer-dependent, so we have to pull them a little differently.
    all_datasources = set(spy.search(
        {'Path': f'{ROOT_NODE} >> {customer} >> {fqdn} >> Datasource', 'Type': 'Asset'},
        recursive=False, quiet=True)['Name'].values)

    eligible_datasources = all_datasources - EXCLUDED_DATASOURCES
    for datasource_class in eligible_datasources:
        path = f'Performance Monitors >> {customer} >> {fqdn} >> Datasource >> {datasource_class}'
        datasource_names = spy.search({'Path': path, 'Type': 'Asset'}, recursive=False, quiet=True)['Name'].values
        for datasource_name in datasource_names:
            calculated_signals_to_trend = []
            signals_to_trend = []
            full_timer_path = f'{path} >> {datasource_name} >> DatasourceTime >> Timer'
            signal_name = '/P95$/'
            timer_result = spy.search({'Path': full_timer_path, 'Name': signal_name}, quiet=True)
            cleaned_name = signal_name.replace('/', '').replace('$', '')
            lane = 1
            if not timer_result.empty:
                signals_to_trend.append(customize_stored_signals_trend_display(timer_result, lane))
                calculated_signals_to_trend.append(
                    create_agile_filter_calculated_signal(f'{full_timer_path} >> {cleaned_name}'[:99],
                                                          timer_result['ID'][0], 1))
                print(f'Created calculated signal {full_timer_path}{cleaned_name}{AGILE_FILTER_5_MIN}')
                lane += 1
            datasource_rx_monitors = [
                'isConnected',
                'Rx >> Bytes',
                'Rx >> Samples',
                'Rx >> Successes',
                'Rx >> Timeouts']
            for monitor in datasource_rx_monitors:
                results = spy.search({'Path': f'{path} >> {datasource_name} >> {monitor}'}, quiet=True)
                if not results.empty:
                    signals_to_trend.append(customize_stored_signals_trend_display(results, lane))
                    calculated_signals_to_trend.append(create_running_delta_agile_filter_calculated_signal(
                        monitor, results['ID'][0], lane))
                lane += 1
            if len(calculated_signals_to_trend) > 0:
                worksheet = workbook.worksheet(f'DS - {datasource_class}.{datasource_name}'[:99])
                worksheet.display_range = {
                    'Start': two_weeks_ago,
                    'End': today
                }
                pushed_calculated_signals_to_trend = push_calculated_signal_metadata(
                    pd.concat(calculated_signals_to_trend),
                    workbook.id
                )
                calculated_signals_to_trend = merge_push_results(calculated_signals_to_trend,
                                                                 pushed_calculated_signals_to_trend)
                set_worksheet_display_items(worksheet, [*signals_to_trend, *calculated_signals_to_trend])
                print(f'Created worksheet for {datasource_class}.{datasource_name}')


def push_workbook(workbook: Analysis,
                  workbook_name: str) -> None:
    url = spy.workbooks.push(
        workbook,
        path=CUSTOMER_MONITORS_FOLDER_ID,
        datasource=DATASOURCE,
        quiet=True
    )['URL'][0]
    print(f'Finished creating {workbook_name}. Go to {url} to access the workbook.')
    spy.acl.push(workbook.id, {
        'ID': SEEQ_PERSONNEL_USER_GROUP_GUID,
        'Read': True,
        'Write': True,
        'Manage': True,
    }, replace=True, disable_inheritance=False, quiet=True)
    print(f'Gave access for {workbook_name} to Seeq Personnel.')


def build_path_str(*path_lists):
    return ' >> '.join(itertools.chain(*path_lists))


def build_name_str(path, name):
    return '.'.join([*path, name])


def variable_digest(data: str, length: int):
    h = shake_128()
    h.update(data.encode())
    return h.hexdigest(length)


def main(
        url: str,
        user: str,
        access_key: str,
        password: str,
        customer: str,
        server: str,
        fqdn: str,
        workbook_name: str,
        overwrite: bool,
        monitoring: str,
) -> None:
    if monitoring not in ['sh', 'ds', 'both']:
        raise 'Provide one of the following for -m flag "ds" for datasources, "sh" for system health or "both"'

    spy.options.allow_version_mismatch = True
    spy.login(url=url, access_key=access_key, username=user, password=password, ignore_ssl_errors=True, quiet=True)
    print("Logged in successfully")
    today = date.today()
    fqdn = fqdn if fqdn else server
    two_weeks_ago = today - datetime.timedelta(days=14)
    workbook_name = workbook_name if workbook_name else f'Monitoring {customer} - {fqdn}'

    if monitoring == 'sh' or monitoring == 'both':
        signal_metadata = spy.search(
            {'Path': f'{ROOT_NODE} >> {customer} >> {fqdn}', 'Type': 'StoredSignal'},
            quiet=True
        )
        if signal_metadata.empty:
            print(
                f'ERROR: {customer}-{fqdn} does not exist. Please make sure the customer and host names are correct.')
            return

        workbook_df = spy.search({'Name': workbook_name, 'Type': 'Workbook'}, quiet=True)
        workbook_df.sort_values('ID')
        if not workbook_df.empty and not overwrite:
            print(
                f'{workbook_name} already exists and -o/--overwrite is false. Please set -o/--overwrite to true if you '
                f'would like to overwrite the existing monitoring workbook.')
            return

        if len(workbook_df) > 0:
            workbook = spy.workbooks.Analysis({'Name': workbook_name, 'ID': workbook_df['ID'][0]})
        else:
            workbook = spy.workbooks.Analysis({'Name': workbook_name})
            workbook.worksheet(list(monitor_group_definitions(server).keys())[0])
            spy.workbooks.push(workbook, quiet=True)

        try:
            create_regular_monitor_worksheets(workbook, signal_metadata, two_weeks_ago, today, server)
        except Exception as e:
            raise Exception(f'Encountered error while building workbook {workbook_name}. No workbook was built {e}')
        push_workbook(workbook, workbook_name)

    if monitoring == 'ds' or monitoring == 'both':
        datasource_workbook_name = workbook_name + DATASOURCES_WORKBOOK_SUFFIX
        workbook_df = spy.search({'Name': datasource_workbook_name, 'Type': 'Workbook'}, quiet=True)
        workbook_df.sort_values('ID')
        if len(workbook_df) > 0:
            workbook_datasources = spy.workbooks.Analysis(
                {'Name': datasource_workbook_name, 'ID': workbook_df['ID'][0]})
        else:
            workbook_datasources = spy.workbooks.Analysis({'Name': datasource_workbook_name})
            workbook_datasources.worksheet(list(monitor_group_definitions(server).keys())[0])
            spy.workbooks.push(workbook_datasources, quiet=True)

        try:
            create_datasource_monitor_worksheets(workbook_datasources, two_weeks_ago, today, customer, fqdn)
        except:
            raise f'Encountered error while building workbook {datasource_workbook_name}. No workbook was built'
        push_workbook(workbook_datasources, datasource_workbook_name)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-u", "--user", required=False, help="User name to log in with")
    parser.add_argument("-p", "--password", required=False, help="Password name to log in with")
    parser.add_argument("-k", "--access_key", required=False, help="Access key to log in with")
    parser.add_argument("-c", "--customer", required=True, help="Customer name for which to build monitoring workbook")
    parser.add_argument("-s", "--server", required=True, help="Customer server name for which to build monitoring "
                                                              "workbook (e.g. xyz-main-p) - case-sensitive")
    parser.add_argument("-f", "--fqdn", required=False,
                        help="Specify a fully-qualified domain name (e.g. xyz.seeq.site) if different from server "
                             "name (xyz-main-p) - case-insensitive, defaults to -s|--server")
    parser.add_argument("-w", "--workbook_name", required=False, help="The name for the monitoring workbook")
    parser.add_argument("-o", "--overwrite", required=False, action='store_true',
                        help="Overwrite the monitoring workbook if the name already exists")
    parser.add_argument("-m", "--monitoring", default="both", required=False,
                        choices=['sh', 'ds', 'both'],
                        help="Create system health (\"sh\"), datasource (\"ds\") or workbooks monitoring both ("
                             "\"both\").")

    args = parser.parse_args()
    MONITORS_URL = "https://monitors.seeq.site/"
    main(MONITORS_URL, args.user, args.access_key, args.password, args.customer, args.server, args.fqdn,
         args.workbook_name, args.overwrite, args.monitoring)


def test_build_path_str():
    assert build_path_str(
        ['Path component 1', 'two'], ['More', 'paths', 'Here'], ['and']
    ) == 'Path component 1 >> two >> More >> paths >> Here >> and'


def test_build_name_str():
    assert build_name_str(
        ['Path component 1', 'two'], 'three'
    ) == 'Path component 1.two.three'



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 59.0.1-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from pprint import pformat
from six import iteritems
import re


class ConditionInputV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'additional_properties': 'list[ScalarPropertyV1]',
        'capsule_properties': 'list[CapsulePropertyInputV1]',
        'data_id': 'str',
        'datasource_class': 'str',
        'datasource_id': 'str',
        'description': 'str',
        'formula': 'str',
        'historical_data_version': 'str',
        'host_id': 'str',
        'maximum_duration': 'str',
        'name': 'str',
        'parameters': 'list[str]',
        'properties': 'list[ScalarPropertyV1]',
        'scoped_to': 'str',
        'security_string': 'str',
        'source_security_string': 'str',
        'sync_token': 'str',
        'unit_of_measure': 'str'
    }

    attribute_map = {
        'additional_properties': 'additionalProperties',
        'capsule_properties': 'capsuleProperties',
        'data_id': 'dataId',
        'datasource_class': 'datasourceClass',
        'datasource_id': 'datasourceId',
        'description': 'description',
        'formula': 'formula',
        'historical_data_version': 'historicalDataVersion',
        'host_id': 'hostId',
        'maximum_duration': 'maximumDuration',
        'name': 'name',
        'parameters': 'parameters',
        'properties': 'properties',
        'scoped_to': 'scopedTo',
        'security_string': 'securityString',
        'source_security_string': 'sourceSecurityString',
        'sync_token': 'syncToken',
        'unit_of_measure': 'unitOfMeasure'
    }

    def __init__(self, additional_properties=None, capsule_properties=None, data_id=None, datasource_class=None, datasource_id=None, description=None, formula=None, historical_data_version=None, host_id=None, maximum_duration=None, name=None, parameters=None, properties=None, scoped_to=None, security_string=None, source_security_string=None, sync_token=None, unit_of_measure=None):
        """
        ConditionInputV1 - a model defined in Swagger
        """

        self._additional_properties = None
        self._capsule_properties = None
        self._data_id = None
        self._datasource_class = None
        self._datasource_id = None
        self._description = None
        self._formula = None
        self._historical_data_version = None
        self._host_id = None
        self._maximum_duration = None
        self._name = None
        self._parameters = None
        self._properties = None
        self._scoped_to = None
        self._security_string = None
        self._source_security_string = None
        self._sync_token = None
        self._unit_of_measure = None

        if additional_properties is not None:
          self.additional_properties = additional_properties
        if capsule_properties is not None:
          self.capsule_properties = capsule_properties
        if data_id is not None:
          self.data_id = data_id
        if datasource_class is not None:
          self.datasource_class = datasource_class
        if datasource_id is not None:
          self.datasource_id = datasource_id
        if description is not None:
          self.description = description
        if formula is not None:
          self.formula = formula
        if historical_data_version is not None:
          self.historical_data_version = historical_data_version
        if host_id is not None:
          self.host_id = host_id
        if maximum_duration is not None:
          self.maximum_duration = maximum_duration
        if name is not None:
          self.name = name
        if parameters is not None:
          self.parameters = parameters
        if properties is not None:
          self.properties = properties
        if scoped_to is not None:
          self.scoped_to = scoped_to
        if security_string is not None:
          self.security_string = security_string
        if source_security_string is not None:
          self.source_security_string = source_security_string
        if sync_token is not None:
          self.sync_token = sync_token
        if unit_of_measure is not None:
          self.unit_of_measure = unit_of_measure

    @property
    def additional_properties(self):
        """
        Gets the additional_properties of this ConditionInputV1.

        :return: The additional_properties of this ConditionInputV1.
        :rtype: list[ScalarPropertyV1]
        """
        return self._additional_properties

    @additional_properties.setter
    def additional_properties(self, additional_properties):
        """
        Sets the additional_properties of this ConditionInputV1.

        :param additional_properties: The additional_properties of this ConditionInputV1.
        :type: list[ScalarPropertyV1]
        """

        self._additional_properties = additional_properties

    @property
    def capsule_properties(self):
        """
        Gets the capsule_properties of this ConditionInputV1.
        Property metadata for capsules on this condition. Only applicable to stored conditions.

        :return: The capsule_properties of this ConditionInputV1.
        :rtype: list[CapsulePropertyInputV1]
        """
        return self._capsule_properties

    @capsule_properties.setter
    def capsule_properties(self, capsule_properties):
        """
        Sets the capsule_properties of this ConditionInputV1.
        Property metadata for capsules on this condition. Only applicable to stored conditions.

        :param capsule_properties: The capsule_properties of this ConditionInputV1.
        :type: list[CapsulePropertyInputV1]
        """

        self._capsule_properties = capsule_properties

    @property
    def data_id(self):
        """
        Gets the data_id of this ConditionInputV1.
        The data ID of this item. Note: This is not the Seeq ID, but the unique identifier that the remote datasource uses.

        :return: The data_id of this ConditionInputV1.
        :rtype: str
        """
        return self._data_id

    @data_id.setter
    def data_id(self, data_id):
        """
        Sets the data_id of this ConditionInputV1.
        The data ID of this item. Note: This is not the Seeq ID, but the unique identifier that the remote datasource uses.

        :param data_id: The data_id of this ConditionInputV1.
        :type: str
        """

        self._data_id = data_id

    @property
    def datasource_class(self):
        """
        Gets the datasource_class of this ConditionInputV1.
        Along with the Datasource ID, the Datasource Class uniquely identifies a datasource. For example, a datasource may be a particular instance of an OSIsoft PI historian.

        :return: The datasource_class of this ConditionInputV1.
        :rtype: str
        """
        return self._datasource_class

    @datasource_class.setter
    def datasource_class(self, datasource_class):
        """
        Sets the datasource_class of this ConditionInputV1.
        Along with the Datasource ID, the Datasource Class uniquely identifies a datasource. For example, a datasource may be a particular instance of an OSIsoft PI historian.

        :param datasource_class: The datasource_class of this ConditionInputV1.
        :type: str
        """

        self._datasource_class = datasource_class

    @property
    def datasource_id(self):
        """
        Gets the datasource_id of this ConditionInputV1.
        Along with the Datasource Class, the Datasource ID uniquely identifies a datasource. For example, a datasource may be a particular instance of an OSIsoft PI historian.

        :return: The datasource_id of this ConditionInputV1.
        :rtype: str
        """
        return self._datasource_id

    @datasource_id.setter
    def datasource_id(self, datasource_id):
        """
        Sets the datasource_id of this ConditionInputV1.
        Along with the Datasource Class, the Datasource ID uniquely identifies a datasource. For example, a datasource may be a particular instance of an OSIsoft PI historian.

        :param datasource_id: The datasource_id of this ConditionInputV1.
        :type: str
        """

        self._datasource_id = datasource_id

    @property
    def description(self):
        """
        Gets the description of this ConditionInputV1.
        Clarifying information or other plain language description of this item. An input of just whitespaces is equivalent to a null input.

        :return: The description of this ConditionInputV1.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this ConditionInputV1.
        Clarifying information or other plain language description of this item. An input of just whitespaces is equivalent to a null input.

        :param description: The description of this ConditionInputV1.
        :type: str
        """

        self._description = description

    @property
    def formula(self):
        """
        Gets the formula of this ConditionInputV1.
        Information about the formula used to create a calculated item. Not marked as required because this class is extended by classes where it is optional

        :return: The formula of this ConditionInputV1.
        :rtype: str
        """
        return self._formula

    @formula.setter
    def formula(self, formula):
        """
        Sets the formula of this ConditionInputV1.
        Information about the formula used to create a calculated item. Not marked as required because this class is extended by classes where it is optional

        :param formula: The formula of this ConditionInputV1.
        :type: str
        """

        self._formula = formula

    @property
    def historical_data_version(self):
        """
        Gets the historical_data_version of this ConditionInputV1.
        This property tells Seeq that certain, and thus potentially cached, capsules in the past have been changed. Changing the value of this property will will clear the cache of this condition and all formulas that depend on it. It should not be changed when new uncached capsules are coming in. Changes to this property have no effect on calculated conditions.

        :return: The historical_data_version of this ConditionInputV1.
        :rtype: str
        """
        return self._historical_data_version

    @historical_data_version.setter
    def historical_data_version(self, historical_data_version):
        """
        Sets the historical_data_version of this ConditionInputV1.
        This property tells Seeq that certain, and thus potentially cached, capsules in the past have been changed. Changing the value of this property will will clear the cache of this condition and all formulas that depend on it. It should not be changed when new uncached capsules are coming in. Changes to this property have no effect on calculated conditions.

        :param historical_data_version: The historical_data_version of this ConditionInputV1.
        :type: str
        """

        self._historical_data_version = historical_data_version

    @property
    def host_id(self):
        """
        Gets the host_id of this ConditionInputV1.
        The ID of the datasource hosting this item. Note that this is a Seeq-generated ID, not the way that the datasource identifies itself.

        :return: The host_id of this ConditionInputV1.
        :rtype: str
        """
        return self._host_id

    @host_id.setter
    def host_id(self, host_id):
        """
        Sets the host_id of this ConditionInputV1.
        The ID of the datasource hosting this item. Note that this is a Seeq-generated ID, not the way that the datasource identifies itself.

        :param host_id: The host_id of this ConditionInputV1.
        :type: str
        """

        self._host_id = host_id

    @property
    def maximum_duration(self):
        """
        Gets the maximum_duration of this ConditionInputV1.
        The maximum duration of capsules in this series. Required for stored conditions.

        :return: The maximum_duration of this ConditionInputV1.
        :rtype: str
        """
        return self._maximum_duration

    @maximum_duration.setter
    def maximum_duration(self, maximum_duration):
        """
        Sets the maximum_duration of this ConditionInputV1.
        The maximum duration of capsules in this series. Required for stored conditions.

        :param maximum_duration: The maximum_duration of this ConditionInputV1.
        :type: str
        """
        if maximum_duration is None:
            raise ValueError("Invalid value for `maximum_duration`, must not be `None`")

        self._maximum_duration = maximum_duration

    @property
    def name(self):
        """
        Gets the name of this ConditionInputV1.
        Human readable name. Required during creation. An input of just whitespaces is equivalent to a null input.

        :return: The name of this ConditionInputV1.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this ConditionInputV1.
        Human readable name. Required during creation. An input of just whitespaces is equivalent to a null input.

        :param name: The name of this ConditionInputV1.
        :type: str
        """
        if name is None:
            raise ValueError("Invalid value for `name`, must not be `None`")

        self._name = name

    @property
    def parameters(self):
        """
        Gets the parameters of this ConditionInputV1.

        :return: The parameters of this ConditionInputV1.
        :rtype: list[str]
        """
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        """
        Sets the parameters of this ConditionInputV1.

        :param parameters: The parameters of this ConditionInputV1.
        :type: list[str]
        """

        self._parameters = parameters

    @property
    def properties(self):
        """
        Gets the properties of this ConditionInputV1.

        :return: The properties of this ConditionInputV1.
        :rtype: list[ScalarPropertyV1]
        """
        return self._properties

    @properties.setter
    def properties(self, properties):
        """
        Sets the properties of this ConditionInputV1.

        :param properties: The properties of this ConditionInputV1.
        :type: list[ScalarPropertyV1]
        """

        self._properties = properties

    @property
    def scoped_to(self):
        """
        Gets the scoped_to of this ConditionInputV1.
        The ID of the workbook to which this item will be scoped.

        :return: The scoped_to of this ConditionInputV1.
        :rtype: str
        """
        return self._scoped_to

    @scoped_to.setter
    def scoped_to(self, scoped_to):
        """
        Sets the scoped_to of this ConditionInputV1.
        The ID of the workbook to which this item will be scoped.

        :param scoped_to: The scoped_to of this ConditionInputV1.
        :type: str
        """

        self._scoped_to = scoped_to

    @property
    def security_string(self):
        """
        Gets the security_string of this ConditionInputV1.
        Security string containing all identities and their permissions for the item. If set, permissions can only be managed by the connector and not in Seeq.

        :return: The security_string of this ConditionInputV1.
        :rtype: str
        """
        return self._security_string

    @security_string.setter
    def security_string(self, security_string):
        """
        Sets the security_string of this ConditionInputV1.
        Security string containing all identities and their permissions for the item. If set, permissions can only be managed by the connector and not in Seeq.

        :param security_string: The security_string of this ConditionInputV1.
        :type: str
        """

        self._security_string = security_string

    @property
    def source_security_string(self):
        """
        Gets the source_security_string of this ConditionInputV1.
        The security string as it was originally found on the source (for debugging ACLs only)

        :return: The source_security_string of this ConditionInputV1.
        :rtype: str
        """
        return self._source_security_string

    @source_security_string.setter
    def source_security_string(self, source_security_string):
        """
        Sets the source_security_string of this ConditionInputV1.
        The security string as it was originally found on the source (for debugging ACLs only)

        :param source_security_string: The source_security_string of this ConditionInputV1.
        :type: str
        """

        self._source_security_string = source_security_string

    @property
    def sync_token(self):
        """
        Gets the sync_token of this ConditionInputV1.
        An arbitrary token (often a date or random ID) that is used during metadata syncs. At the end of a full sync, items with mismatching sync tokens are identified as stale and may be archived using the Datasources clean-up API.

        :return: The sync_token of this ConditionInputV1.
        :rtype: str
        """
        return self._sync_token

    @sync_token.setter
    def sync_token(self, sync_token):
        """
        Sets the sync_token of this ConditionInputV1.
        An arbitrary token (often a date or random ID) that is used during metadata syncs. At the end of a full sync, items with mismatching sync tokens are identified as stale and may be archived using the Datasources clean-up API.

        :param sync_token: The sync_token of this ConditionInputV1.
        :type: str
        """

        self._sync_token = sync_token

    @property
    def unit_of_measure(self):
        """
        Gets the unit_of_measure of this ConditionInputV1.

        :return: The unit_of_measure of this ConditionInputV1.
        :rtype: str
        """
        return self._unit_of_measure

    @unit_of_measure.setter
    def unit_of_measure(self, unit_of_measure):
        """
        Sets the unit_of_measure of this ConditionInputV1.

        :param unit_of_measure: The unit_of_measure of this ConditionInputV1.
        :type: str
        """

        self._unit_of_measure = unit_of_measure

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, ConditionInputV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



from __future__ import annotations

"""
  Implements utility methods used to determine information and run commands across different platforms.
"""
import os
import pathlib
import re
import threading
from typing import Any, AnyStr, Iterable, IO, Mapping, Optional, Tuple, Union


_AnyPath = Union["os.PathLike[AnyStr]", AnyStr]

# `spawn`'s file handle type (stdin, stdout, stderr)
# This is analogous to the corresponding `_FILE` type in Popen's type stub.
_FILE = Union[None, int, IO[Any]]

# `spawn`'s command type. This is analogous to the corresponding `_CMD` type in Popen's stub.
SpawnCommand = Union[_AnyPath, Iterable[_AnyPath]]


# See http://code.activestate.com/recipes/52308-the-simple-but-handy-collector-of-a-bunch-of-named
class Bunch:
    def __init__(self, **kwds):
        self.__dict__.update(kwds)


UNITS_BASE_TWO = ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB"]
UNITS_BASE_TEN = ["B", "KB", "MB", "GB", "TB", "PB", "EB"]
BYTES_PER_KIBIBYTE = 1024
BYTES_PER_MEBIBYTE = BYTES_PER_KIBIBYTE * BYTES_PER_KIBIBYTE
BYTES_PER_MEGABYTE = 1e6
BYTES_PER_GIGABYTE = 1e9


def get_system():
    import platform
    return platform.system()


def get_release():
    import platform
    return platform.release()


def get_architecture():
    import platform
    return platform.architecture()


def get_platform():
    import platform
    plat = platform.platform()
    if plat.lower().find('windows') != -1:
        return 'windows'
    if plat.lower().find('linux') != -1:
        return 'linux'
    if plat.lower().find('darwin') != -1 or plat.lower().find('macos') != -1:
        return 'osx'


def get_linux_distribution():
    import platform
    # Examples of platform.platform() return values:
    # Ubuntu 18.10 on Azure:  Linux-5.0.0-1027-azure-x86_64-with-debian-buster-sid

    if not is_linux():
        return None

    platform_map = {
        'debian': 'debian',
        'ubuntu': 'debian',
        'fedora': 'fedora',
        'redhat': 'fedora',
        'centos': 'fedora'
    }

    plat = platform.platform().lower()
    for fragment, distribution in platform_map.items():
        if fragment in plat:
            return distribution

    return None


def get_system_bit():
    if is_64_bit():
        return '64bit'
    else:
        return '32bit'


def get_build_architecture():
    return os.environ['SQ_ARCHITECTURE']


def is_build_x86():
    return get_build_architecture() == 'x86'


def file_empty(path):
    apath = os.path.join(get_project_root_dir(), path)
    return not os.path.isfile(apath) or os.stat(apath).st_size == 0


def print_file(path, indent=''):
    num_lines = 0
    with open(os.path.join(get_project_root_dir(), path)) as f:
        for line in f.readlines():
            print((indent + line))
            num_lines += 1
    return num_lines


def is_windows():
    return get_platform() == 'windows'


def is_linux():
    return get_platform() == 'linux'


def is_osx():
    return get_platform() == 'osx'


def is_osx_aarch64():
    import platform
    return get_platform() == 'osx' and platform.platform().lower().find('arm64') != -1


g_repo_root_dir = None


def get_repo_root_dir(windows_long_path=True):
    r"""
    Returns the path to the root of the repository. i.e., C:\crab\
    """
    global g_repo_root_dir
    if not g_repo_root_dir:
        # Cache it since this is an expensive call
        dir1 = os.path.abspath(os.path.dirname(__file__))
        while not os.path.exists(os.path.join(dir1, 'repo-root-dir.txt')):
            dir2 = os.path.abspath(os.path.join(dir1, os.pardir))
            if dir1 == dir2:
                # We hit the root and didn't find it
                return

            dir1 = dir2

        g_repo_root_dir = dir1

    return cleanse_path(g_repo_root_dir, windows_long_path)


def get_git_dir(windows_long_path=True):
    r"""
    In the most common case, returns the path to the to the repository's
    '.git' folder. i.e., C:\crab\.git
    """
    return cleanse_path(os.path.abspath(git_rev_parse('--git-dir')), windows_long_path)


def get_git_common_dir(windows_long_path=True):
    """
    Returns the path to the '.git' folder. In git worktrees the
    '--git-dir' flag returns a path to a text file. In a worktree
    the '--git-common-dir' flag returns the path to the true
    '.git' directory in the main git worktree
    """
    # noinspection PyBroadException
    try:
        # common dir is incorrect when ran from a sub folder
        # See https://goo.gl/MTHFQf
        wd = get_repo_root_dir(windows_long_path)
        return cleanse_path(os.path.join(wd, git_rev_parse('--git-common-dir', cwd=wd)), windows_long_path)
    except BaseException:
        # We are likely on an older version of git, use the --git-dir flag (which has existed longer)
        return get_git_dir(windows_long_path)


def get_project_name():
    project_root_dir = get_project_root_dir()

    # Get rid of everything before the last forward slash
    return project_root_dir.split(os.sep)[-1]


g_project_root_dir = ''


def get_project_root_dir(windows_long_path=True):
    if g_project_root_dir == '':
        raise Exception('g_project_root_dir not set')

    project_root_dir = cleanse_path(g_project_root_dir, windows_long_path)

    return project_root_dir


def set_project_root_dir(project_root_dir):
    global g_project_root_dir
    g_project_root_dir = project_root_dir


def get_common_dir(windows_long_path=True):
    common_dir = os.path.join(g_project_root_dir, '..', 'common')
    common_dir = os.path.normpath(common_dir)

    if not os.path.exists(common_dir):
        common_dir = os.path.join(g_project_root_dir, 'common')

    if not os.path.exists(common_dir):
        raise Exception('common directory could not be found')

    return cleanse_path(common_dir, windows_long_path)


def get_home_dir(user=None):
    if is_windows():
        home = os.environ['USERPROFILE']
    else:
        user_to_expand = user if user else ''
        home = os.path.expanduser(f'~{user_to_expand}')

    return home


def get_sq_dir():
    return os.environ['SQ_FOLDER']


def is_64_bit():
    if is_windows():
        # On Windows, even if Python is running as a 32-bit process, we are actually
        # more interested in whether we are on a 64-bit flavor of Windows.
        # In a 32-bit process on 64-bit Windows, PROCESSOR_ARCHITEW6432 will be AMD64.
        # In a 64-bit process on 64-bit Windows, PROCESSOR_ARCHITECTURE will be AMD64.
        return os.environ.get('PROCESSOR_ARCHITEW6432') == 'AMD64' or os.environ.get(
            'PROCESSOR_ARCHITECTURE') == 'AMD64'
    else:
        import struct
        # On UNIX operating systems, Python will be either 32- or 64-bit according to the
        # operating system. So we can just look at the size of a pointer.
        return True if struct.calcsize("P") * 8 == 64 else False


def is_os_headless():
    """
      Returns True if a display is not detected in the build environment
    """
    return get_platform() == 'linux' and 'DISPLAY' not in os.environ


def render_command_line_posix(cmd_and_args: Iterable[Union[str, bytes]]) -> str:
    """
    Render a command line from its program + argument list into a form that can be consumed by a POSIX-like
    shell (sh, bash, zsh, etc) in which each argument is preserved unambiguously.

    >>> print(render_command_line_posix(['find', 'Cool Beans/', '-iname', '*.txt']))
    find 'Cool Beans/' -iname '*.txt'
    """
    import shlex
    return ' '.join(shlex.quote(os.fsdecode(word)) for word in cmd_and_args)


def render_command_line_msvcrt(cmd_and_args: Iterable[Union[str, bytes]]) -> str:
    """
    Render a command line from its program + argument list into a form that can be interpreted by the MSVCRT on Windows,
    for which each argument is preserved unambiguously.

    Note that this does NOT suppress cmd.exe-specific syntax that might be applied prior to execution, if run that way,
    e.g. environment variable expansions: it just surrounds tokens by quotes as necessary to group them.

    >>> print(render_command_line_msvcrt(['dir', '/q', 'Some Folder']))
    dir /q "Some Folder"
    """
    from subprocess import list2cmdline
    return list2cmdline([os.fsdecode(word) for word in cmd_and_args])


def render_command_line(cmd_and_args: Iterable[Union[str, bytes]]) -> str:
    """
    Render a command line from its program + argument list for the current platform.
    The quoting is performed as a command for `cmd.exe` on Windows, or a POSIX-ish shell otherwise.

    See also `render_command_line_msvcrt` and `render_command_line_posix`.
    """
    if is_windows():
        return render_command_line_msvcrt(cmd_and_args)
    else:
        return render_command_line_posix(cmd_and_args)


def spawn(
        command_line: SpawnCommand,
        *,
        raise_on_error: bool = True,
        capture_output: bool = False,
        wait_for_end: bool = True,
        suppress_input: bool = False,
        suppress_output: bool = False,
        suppress_error: bool = False,
        cwd: _AnyPath = None,
        env: Mapping[str, str] = None,
        echo: bool = False,
        shell: Optional[bool] = None,
        stdout: _FILE = None,
        stderr: _FILE = None,
):
    """
      Runs the 'command_line' input in the platform's command line.
    """
    import subprocess
    from subprocess import PIPE
    import sys

    if not cwd:
        # noinspection PyBroadException
        try:
            cwd = get_project_root_dir(windows_long_path=False)
        except BaseException:
            cwd = os.getcwd()
    elif sys.version_info < (3, 8):
        # Remove me when Python 3.7 support is dropped: `Popen` does this itself in 3.8 and above.
        cwd = os.fsdecode(cwd)

    if env is None:
        env = os.environ

    stdin = None

    if not stderr:
        # By default, we pipe stderr into stdout because Bamboo buffers stdout and not stderr, so you
        # can get out-of-order logs that are really confusing.
        # See https://answers.atlassian.com/questions/31595/bamboo-build-log-not-in-sync
        stderr = subprocess.STDOUT

    output = ''
    err = ''

    if capture_output and suppress_output:
        raise Exception('capture_output and suppress_output are mutually exclusive.')

    if capture_output:
        stdout = PIPE
        stderr = PIPE

    if suppress_input:
        stdin = open(os.devnull, 'r')

    if suppress_output:
        stdout = open(os.devnull, 'w')

    if suppress_error:
        stderr = open(os.devnull, 'w')

    if isinstance(command_line, (str, bytes)):
        command_line = os.fsdecode(command_line)
    elif hasattr(command_line, '__fspath__'):
        # We have a single PathLike: treat it as a program/script name.
        command_line = (os.fsdecode(command_line),)
    else:
        # Make sure to consume the arguments exactly once, in case they come from an iterator (e.g. a generator).
        # Convert any internal PathLikes, while we're at it.
        command_line = tuple(os.fsdecode(word) for word in command_line)

    if shell is None:
        # Assume that if the command is a string, then it's a pre-formatted command line to be interpreted by
        # cmd.exe or sh. Otherwise run as a proper exec command ([program, argv1, argv2, …]).
        # Although you *can* use the shell with an argument list, that is a slightly more advanced mode of operation,
        # and behaves wildly differently for windows vs. posix, and so you should pass `shell=True` explicitly
        # if you want that.
        shell = isinstance(command_line, str)

    if echo:
        if is_windows():
            if isinstance(command_line, str):
                cmd_str = command_line
            else:
                cmd_str = render_command_line_msvcrt(command_line)

            if shell:
                comspec = os.environ.get('COMSPEC', 'cmd.exe')
                cmd_str = f'{comspec} /c "{cmd_str}"'
        else:
            if isinstance(command_line, str):
                cmd_args = (command_line,)
            else:
                cmd_args = command_line

            if shell:
                cmd_args = ('/bin/sh', '-c', *cmd_args)
            cmd_str = render_command_line_posix(cmd_args)
        print(cmd_str)

    proc = subprocess.Popen(
        command_line,
        shell=shell,
        env=env,
        cwd=cwd,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr,
    )

    if capture_output:
        (output, err) = proc.communicate()

    def decode(s, encodings):
        if not isinstance(s, str):
            for encoding in encodings:
                try:
                    return str(s, encoding=encoding, errors='replace')
                except UnicodeDecodeError:
                    pass
            return str(s)
        else:
            return s

    if wait_for_end:
        ret_code = proc.wait()

        if ret_code != 0 and raise_on_error:
            raise RuntimeError('Command failed with exit code %d: %s' %
                               (ret_code, command_line))

        output = decode(output, encodings=(sys.stdout.encoding, 'ascii', 'utf-8', 'latin1', 'windows-1252'))
        err = decode(err, encodings=(sys.stderr.encoding, 'ascii', 'utf-8', 'latin1', 'windows-1252'))

        return ret_code, output, err

    else:
        return proc


def git_rev_parse(flag, cwd=None):
    """
    Calls the 'git rev-parse' command with the supplied flag. This is useful for
    getting a path from git. For example, 'git rev-parse --git-dir' will return
    the directory of the '.git' directory.
    """
    (ret_code, output, err) = spawn('git rev-parse --no-flags ' + flag, capture_output=True, cwd=cwd)
    result = output.strip()
    if result == '':
        raise Exception('git rev-parse failed to return a path')
    return result


def convert_windows_path_to_unix(path):
    return re.sub(r'\\', '/', path)


def cleanse_path(path: Union[str, pathlib.Path], windows_long_path: bool = True):
    """
    Takes a potentially 'mixed-platform' path (i.e., partially Windows, partially Unix)
    and turns it into a path that is guaranteed to work on the host platform. Also,
    on Windows it will optionally turn it into a 'long path' that can support >65k
    characters. (This is necessary especially for doing any operations on the
    node_modules folder.)
    """

    # Make sure it is a string (could have been a pathlib.Path)
    path = str(path)

    # Get rid of extra/unnecessary /../ or /./ sections of the path. Note that normpath does not work properly with
    # long filenames, so we have to convert to short paths and then call this before cleaning.
    path = os.path.normpath(re.sub(r'^([/\\][/\\]\?[/\\])?(\w:)', r'\2', path))

    if is_windows():
        # First replace all forward slashes with backslashes like Windows wants.
        path = re.sub(r'/', r'\\', path)
        if windows_long_path:
            # Windows 'long paths' are of the form \\?\D:\Foo\Bar (instead of the
            # regular D:\Foo\Bar). So we have to do a RegEx replacement that involves
            # A LOT of backslashes.
            path = re.sub(r'^(\w):\\', r'\\\\?\\\1:\\', path)
        else:
            # Get rid of the 'long path' form if it's in there. There are a some
            # APIs/utilities that can't handle them.
            path = re.sub(r'^(\\\\\?\\)?(\w:)', r'\2', path)
    else:
        path = convert_windows_path_to_unix(path)

    return path


def cleanse_filename(filename, replacement_char='_'):
    return re.sub(r'[:"%\\/<>^|?*&\[\]]', replacement_char, filename)


def concat_path(path1, path2):
    return cleanse_path(path1 + os.sep + path2)


def script_path_to_unix(path, retain_colon=False):
    r"""
    This function is used to convert a path into something that is suitable for
    a Unix script, including a Bash script being run on Windows in Git Bash.

    On Windows, a path like r"C:\Foo\Bar" or r"\\?\C:\Foo\Bar" will get turned into
    "/c/foo/bar".
    """
    path = path.replace('\\', '/')
    path = re.sub(r'^(//\?/)?(\w):', '/\\2%s' % (':' if retain_colon else ''), path)
    return path


def script_path_to_windows(path):
    """
    This function is used to convert a path into something that is suitable for
    a Windows batch script.

    On Windows, a path like "C:/Foo/Bar"  will get turned into
    'C:\foo\bar'. Note that the long path prefix ('\\?\') will also get stripped
    off because it's not supported in Windows batch scripts. (Go figure.)
    """
    path = path.replace('/', '\\')
    path = re.sub(r'^(\\\\\?\\)?(\w:)', '\\2', path)
    return path


def path_to_regex(path):
    """
    This function returns a regular expression that will match the given path
    properly against a string that may contain a variety of representations of
    the path on Windows.
    """
    # On Windows, paths may be prefixed in the following ways:
    # Regular:      C:\
    # Long Paths:   \\?\C:\
    # Git Bash:     /C/

    # The following RegEx extracts the drive letter no matter which of the
    # ways happens to be used
    *drive_match, remaining_path = re.split(r'^(?:(?:\\\\\?\\)?|/)(\w)(?::[\\/]|/)', os.fsdecode(path), maxsplit=1)

    # We use an almost identical RegEx for our pattern except we hard-code
    # the drive letter, and we do not anchor it to the start of the string.
    prefix_re = fr'(?:(?:\\\\\?\\)?|/){re.escape(drive_match[1])}(?::[\\/]|/)' if drive_match else ''

    # Windows is loosey-goosey in that it allows a mixture of backslashes and
    # forward slashes throughout the path, so we replace all of them with
    # a character matching set that includes both
    path_re = r'[\\/]'.join(map(re.escape, re.split(r'[\\/]', remaining_path)))

    # The path could be wrapped in quotes
    # nit: rather, an *argument* referring to this path could be, and that's what these regexes are being used for.
    quote_re = '[\'"]?'

    # Join the component patterns together.
    return str().join((quote_re, prefix_re, path_re, quote_re))


def find_files_recursively(starting_dir, wildcard):
    import fnmatch
    """
    Returns a list of files that match the wildcard passed in (honoring asterisks)
    by recursively searching the starting_dir and below.
    """
    matches = []
    for root, dirnames, filenames in os.walk(starting_dir):
        for filename in fnmatch.filter(filenames, wildcard):
            matches.append(cleanse_path(os.path.join(root, filename), False))
    return matches


def find_folders_recursively(starting_dir, wildcard):
    import fnmatch
    """
    Returns a list of folders that match the wildcard passed in (honoring asterisks)
    by recursively searching the starting_dir and below.
    """
    matches = []
    for root, dirnames, filenames in os.walk(starting_dir):
        for folder in fnmatch.filter(dirnames, wildcard):
            matches.append(cleanse_path(os.path.join(root, folder), False))
    return matches


def replace_in_string(_str, replacements):
    """
    Replaces all instances of a regex in a string with the specified replacement.
    'replacements' is a list of tuples in the form (pattern, repl, [flags]) where 'pattern'
    is a regex pattern and 'repl' is either a function that, when called with the
    match object returns the replacement text, or a straight regex string that
    represents the replacement. 'flags' is an optional modifier for the regex
    such as re.MULTILINE
    """

    from inspect import isfunction
    while True:
        original = _str
        for replacement in replacements:
            pattern = replacement[0]
            repl = replacement[1]
            flags = replacement[2] if len(replacement) == 3 else 0
            compiled = re.compile(pattern, flags)
            match = compiled.search(_str)
            if match:
                if isfunction(repl):
                    replacement_text = repl(match)
                    _str = _str[:match.start()] + replacement_text + _str[match.end():]
                else:
                    _str = re.sub(pattern, repl, _str, 0, flags)

        if original == _str:
            break

    return _str


def replace_in_file(file_path, replacements):
    """
    Replaces all instances of a regex in a text file with the specified replacement.
    'replacements' is a list of tuples in the form (pattern, repl, [flags]) where 'pattern'
    is a regex pattern and 'repl' is either a function that, when called with the
    match object returns the replacement text, or a straight regex string that
    represents the replacement. 'flags' is an optional modifier for the regex
    such as re.MULTILINE
    """
    file = open(file_path, 'r', encoding='utf-8')
    file_content = file.read()
    file.close()

    file_content = replace_in_string(file_content, replacements)

    file = open(file_path, 'w', encoding='utf-8')
    file.write(file_content)
    file.close()


def replace_in_binary_file(file_path, replacements):
    """
    Replaces all instances of a pattern in a binary file with the specified replacement.
    'replacements' is a list of tuples in the form (pattern, repl) where 'pattern'
    is a pattern and 'repl' is its replacement. Regex is not supported.
    """
    file = open(file_path, 'rb')
    file_content = file.read()
    file.close()

    for replacement in replacements:
        (pattern, repl) = replacement
        file_content = file_content.replace(pattern, repl)

    file = open(file_path, 'wb')
    file.write(file_content)
    file.close()


def get_username():
    """
    Returns the current user's name in a cross platform way.
    """
    if is_windows():
        import win32api
        return win32api.GetUserName()
    else:
        import pwd
        return pwd.getpwuid(os.getuid()).pw_name


def keypress_event():
    """
    Returns a threading.Event object that is signaled if the user presses ENTER.
    """
    event = threading.Event()
    threading.Thread(target=_wait_for_keypress_thread, daemon=True, args=(event,)).start()
    return event


def _wait_for_keypress_thread(event):
    try:
        input()
    except EOFError:
        # If this occurs, then there is no input stream to read from. We exit this thread
        # and the original thread will wait forever (or until its timeout) for the event
        # to be set.
        #
        # This is the case for 'sq run' during 'sq build' for AppServer: The integration
        # test target manager kills 'sq run' using 'sq run --kill', not a keypress.
        return

    event.set()


def is_script_running_on_the_continuous_integration_server():
    return 'IS_CI' in os.environ


def copytree(src, dst, symlinks=False, update=False, mirror=False, permissions=False, progress=False, exclude=None):
    if is_windows():
        robocopy(src, dst, update=update, mirror=mirror, permissions=permissions, progress=progress, exclude=exclude)
    else:
        rsync(src, dst, symlinks=symlinks, update=update, mirror=mirror, permissions=permissions, progress=progress,
              exclude=exclude)


def copyfile(src, dst):
    import stat
    import shutil
    if not os.path.exists(os.path.dirname(dst)):
        os.makedirs(os.path.dirname(dst))

    try:
        shutil.copy2(src, dst)
    except IOError as e:
        if e.errno == 13:  # Permission denied
            # Remove the read-only flag and try again
            os.chmod(dst, stat.S_IWRITE)
            shutil.copy2(src, dst)
        else:
            raise e


def on_rm_error(func, path, exc_info):
    # From
    # http://stackoverflow.com/questions/4829043/how-to-remove-read-only-attrib-directory-with-python-in-windows

    # path contains the path of the file that couldn't be removed
    # let's just assume that it's read-only and unlink it.
    import stat
    os.chmod(path, stat.S_IWRITE)
    os.unlink(path)


def copy_windows_acls(src, dst):
    import tempfile
    import codecs
    # From https://confidentialfiles.wordpress.com/2014/03/13/copying-ntfs-permissions-between-folders/

    # Create a temporary file for the ACL descriptors.
    temp_handle, temp_path = tempfile.mkstemp()
    os.close(temp_handle)

    # Write the ACL descriptors to the temp file.
    spawn('icacls "%s" /save "%s"' % (src, temp_path))

    # Open up the ACL descriptor file. It is encoded as UCS-2 Little Endian.
    with codecs.open(temp_path, 'r', 'utf_16_le') as f:
        lines = f.readlines()

    # Replace the first line, which is the name of the source directory, with the destination directory.
    lines[0] = str(os.path.basename(dst))

    # Write out the new ACL descriptor file.
    with codecs.open(temp_path, 'w', 'utf_16_le') as f:
        f.writelines(os.linesep.join(lines) + os.linesep)

    # Grab the directory name of the destination because (and this is so cool -- not!!) icacls restores ACLs by
    # using the parent directory as the command line input and then the target directory on the first line of the file.
    dst_dir = os.path.dirname(dst)

    # What a pain in the butt! icacls chokes if you pass in "C:\" (with surrounding quotes) for root folders.
    # So only wrap in quotes if there is actually a space somewhere in the directory name.
    if ' ' in dst_dir:
        dst_dir = '"%s"' % dst_dir

    # Finally, "restore" the ACLs to the new location
    spawn('icacls %s /restore "%s"' % (dst_dir, temp_path))

    # Clean up the temp file.
    os.remove(temp_path)


def robocopy(src, dst, mirror=False, update=False, permissions=False, progress=False, exclude=None):
    # Arguments cannot use long path names
    src = cleanse_path(src, windows_long_path=False)
    dst = cleanse_path(dst, windows_long_path=False)
    mode = '/MIR' if mirror else '/E'
    upd = '/XO' if update else ''
    logging = '/NP /NS /NC /NFL /NDL /NJH /NJS'
    if progress:
        logging = '/NP /NS /NC /NJH /NJS'
    exclusions = ''
    if exclude is not None:
        files, dirs = map(lambda ps: ' '.join(ps), _partition(
            lambda path: os.path.isdir(path),
            # get all the existing files that match the exclude path on the src or dst
            (os.path.join(base, x) for x in exclude for base in [src, dst] if os.path.exists(os.path.join(base, x)))
        ))
        if len(files):
            exclusions += ' /XF ' + files
        if len(dirs):
            exclusions += ' /XD ' + dirs
    cmdline = 'robocopy "%s" "%s" %s %s /R:0 /MT %s %s' % (src, dst, mode, upd, logging, exclusions)
    (ret_code, output, err) = spawn(cmdline, capture_output=not progress, raise_on_error=False)
    if ret_code > 3:
        raise Exception(
            'Robocopy failed to copy "%s" to "%s" with error code %d:\n%s\n%s' % (src, dst, ret_code, output, err))

    # Why not use the /SEC switch on robocopy? Because it doesn't seem to work in Windows Server 2008 R2. CRAB-12017
    if permissions:
        copy_windows_acls(src, dst)


def _partition(pred, iterable):
    # Taken from itertools-more (MIT Licensed). It is problematic to add dependencies to seeq.base because they wont
    # be installed early enough and it would be an additional required dependency for spy
    # https://more-itertools.readthedocs.io/en/stable/_modules/more_itertools/recipes.html#partition

    import itertools
    t1, t2 = itertools.tee(((pred(x), x) for x in iterable))
    return (
        (x for (cond, x) in t1 if not cond),
        (x for (cond, x) in t2 if cond),
    )


def rsync(src, dst, symlinks=False, mirror=False, update=False, permissions=False, progress=False, exclude=None):
    if not os.path.exists(dst):
        os.makedirs(dst)

    args = '-rtE'
    if symlinks:
        args = args + 'l'

    if update:
        args = args + 'u'

    if permissions:
        args = args + 'p'

    if not progress:
        args = args + 'q'

    if mirror:
        args = args + ' --del'

    if exclude is not None:
        args = args + ' ' + ' '.join([('--exclude %s' % x) for x in exclude])

    # Add a trailing slash so that rsync won't create an additional level below the destination
    if not src.endswith('/'):
        src = src + '/'

    cmdline = 'rsync %s "%s" "%s"' % (args, src, dst)
    spawn(cmdline)


def do_with_retry(func, timeout_sec=5, retry_sleep_sec=0.1):
    import time
    timer = time.time()
    while True:
        # noinspection PyBroadException
        try:
            func()
            break
        except BaseException:
            if time.time() - timer > timeout_sec:
                raise

            time.sleep(retry_sleep_sec)


def removetree(path, keep_top_folder=False, exclude_subdirectories=None):
    import shutil
    import tempfile
    if exclude_subdirectories is None:
        exclude_subdirectories = list()

    with tempfile.TemporaryDirectory() as empty_folder:
        # If we're on Windows, and there is a folder with contents where part of the path >260 characters, we must
        # use Robocopy (via copytree) to remove the contents of the folder due to extremely long path names.
        # Ironically, however, the top-level folder can't be long-named.
        copytree(str(empty_folder), path, mirror=True, exclude=exclude_subdirectories)

    if not keep_top_folder and len(exclude_subdirectories) == 0:
        # the robocopy process can hold on to the directory we're trying to delete for a short time.
        do_with_retry(lambda: shutil.rmtree(path, onerror=on_rm_error))


def remove_deep_folder(path):
    removetree(path)


def remove_files_from_list(file_list: list):
    for file_path in file_list:
        try:
            os.remove(file_path)
        except OSError:
            print("Error while deleting file : ", file_path)


def get_git_branch() -> Optional[str]:
    """
    Return the name of the current git branch.

    Taken from:
    http://stackoverflow.com/questions/1593051/how-to-programmatically-determine-the-current-checked-out-git-branch
    :return: Name of the current git branch or None if currently in a 'detached HEAD' state
    """
    ret_code, output, err = spawn('git symbolic-ref HEAD', capture_output=True, raise_on_error=False)
    if ret_code != 0:
        if 'ref HEAD is not a symbolic ref' in err:
            return None  # Detached head, no branch
        message = err.replace('fatal: ', '').strip()
        raise RuntimeError(f'Unable to get git branch: {message}')
    return output.replace('refs/heads/', '').strip()


def get_aws_metadata(metadata):
    """
    Return information about the AWS instance that this code is running on.

    Taken from:
    http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/ec2-instance-metadata.html
    :param metadata: The key for the metadata you want to retrieve
    :return: The desired metadata
    """
    import urllib.request
    import urllib.parse
    url = 'http://169.254.169.254/latest/meta-data/%s' % metadata
    with urllib.request.urlopen(url, timeout=1) as f:
        return f.read().decode('utf-8')


def get_current_user_name_and_email() -> Tuple[Optional[str], Optional[str]]:
    user_name = None
    user_email = None

    # noinspection PyBroadException
    try:
        (ret_code, output, err) = spawn('git config user.name', capture_output=True)
        user_name = output.strip() if len(output.strip()) > 0 else None
        (ret_code, output, err) = spawn('git config user.email', capture_output=True)
        user_email = output.strip() if len(output.strip()) > 0 else None
    except BaseException:
        # Didn't work, pass back None
        pass

    return user_name, user_email


def prepend_to_path_environment_variable(env, path_to_prepend):
    # On UNIX, it's PATH, on Windows it's Path
    variable_name = 'PATH'
    if 'Path' in env:
        variable_name = 'Path'

    path_variable = env[variable_name]

    path_variable = path_to_prepend + os.pathsep + path_variable

    env[path_variable] = path_variable

    return env


def get_md5_hash(file_name):
    import hashlib
    with open(file_name, 'rb') as file_to_check:
        data = file_to_check.read()
        return hashlib.md5(data).hexdigest()


def is_admin():
    import ctypes
    try:
        return os.getuid() == 0
    except AttributeError:
        return ctypes.windll.shell32.IsUserAnAdmin()


def fix_permissions(file_or_folder):
    if is_windows():
        # If this is Windows, we might be in an elevated process and that means we'll create the folder with
        # restricted access, which can cause problems later. Use icacls to give "Everyone" access. Administrators
        # can lock it down later.
        cmdline = 'icacls "%s" /grant *S-1-1-0:%sF >NUL' % (file_or_folder,
                                                            '(OI)(CI)' if os.path.isdir(file_or_folder) else '')
        spawn(cmdline)
    else:
        if os.path.isdir(file_or_folder):
            cmdline = f'chmod -R g=rwx,o=rwx "{file_or_folder}"'
        else:
            cmdline = f'chmod g=rw,o=rw "{file_or_folder}"'
        spawn(cmdline)


def create_folder_if_necessary_with_correct_permissions(folder):
    if not os.path.exists(folder):
        os.makedirs(folder)
        fix_permissions(folder)


def force_unicode(s):
    if isinstance(s, str):
        return s
    elif isinstance(s, bytes):
        return str(s, errors='replace')
    else:
        return str(s)


def force_print(s):
    if s is None:
        return

    print(force_unicode(s))


def get_branch_details():
    (ret_code, output, err) = spawn('git status', capture_output=True)

    if ret_code != 0:
        raise RuntimeError('Error calling "git status"')

    branch_name = None
    for match in re.finditer(r'On branch (.*?)\n', output):
        branch_name = match.group(1)

    if not branch_name:
        raise RuntimeError('ERROR: Could not discern git branch from "git status":\n%s' % output)

    pushed = False
    if 'Your branch is' in output:
        pushed = True
    else:
        # The local branch isn't tracking a remote one, but it may still have been pushed. Query directly.
        (ret_code_new, output_new, err_new) = spawn('git ls-remote origin ' + branch_name, capture_output=True)
        if branch_name in output_new:
            pushed = True

    matcher = re.match(r'.*crab-(\d+).*', branch_name.lower())
    if not matcher:
        return branch_name, None, pushed

    crab_number = matcher.group(1)

    return branch_name, crab_number, pushed


def human_readable_byte_count(bytes, use_base_two, round_to_int):
    import math
    """
    Python version of SystemInfo#humanReadableByteCount.
    Constructs a human readable string of the given bytes, such as '2.02GiB' or '9.75TB'.
    :param bytes: Number of bytes
    :param use_base_two: If true, use base-two units (GiB); if false, use base-ten (GB)
    :param round_to_int: If true, round the result to the nearest integer. If false, return the result rounded to two
    decimal points.
    :return: A human readable string of the given bytes, such as '2.02GiB' or '9.75TB'.
    """
    base_kilo = 1024 if use_base_two else 1000

    if bytes < base_kilo:
        return "%s B" % bytes

    exponent = int(math.log(bytes) // math.log(base_kilo))
    number = bytes / (base_kilo ** exponent)

    if round_to_int and (round(number) >= base_kilo):
        exponent += 1
        number = bytes / base_kilo ** exponent

    units = UNITS_BASE_TWO[exponent] if use_base_two else UNITS_BASE_TEN[exponent]

    if round_to_int:
        return "%d %s" % (round(number), units)
    else:
        return "%.2f %s" % (number, units)


def pytest(mark, cwd, parallelize=False, echo=False):
    parallel_arg = '-n auto' if parallelize else ''
    spawn(f'pytest -v -m {mark} --timeout 3600 {parallel_arg}', cwd=cwd, echo=echo)


def periodic_background_task(period: int, initial_delay=0, times=-1):
    """
    Annotate a function with e.g. `periodic_background_task(period=30, initial_delay=5)` to execute that function every
    30 seconds until the python script exits. The task is executed on a daemon thread created just for the task.

    :param period: Seconds between task executions as measured from the end of one execution to the start of the next.
    :param initial_delay: Seconds to delay the first task execution.
    :param times: Number of times to execute the task. -1 to execute forever.
    """

    def outer_wrap(function):
        def wrap(*args, **kwargs):
            stop = threading.Event()

            def inner_wrap():
                i = 0
                stop.wait(initial_delay)
                while i != times and not stop.isSet():
                    function(*args, **kwargs)
                    stop.wait(period)
                    i += 1

            t = threading.Timer(0, inner_wrap)
            t.daemon = True
            t.start()
            return stop

        return wrap

    return outer_wrap


def unpack(file_to_unpack, destination, is_targz=True):
    """ Unpack a .zip file or a .tar.gz file to a given destination
    :param file_to_unpack: String - a path to a .zip file or .tar.gz
    :param destination: String - a path to the destination directory
    :param is_targz: boolean - true if .tar.gz and false if .zip
    """
    import tarfile
    import zipfile
    print('Unpacking %s to %s' % (file_to_unpack, destination))
    if is_targz:
        t = tarfile.open(file_to_unpack, 'r:gz')
        t.extractall(destination)
        t.close()
    else:
        z = zipfile.ZipFile(file_to_unpack, "r")
        z.extractall(destination)
        z.close()
    os.remove(file_to_unpack)



from __future__ import annotations

from enum import Enum
from typing import Optional, Dict

import pandas as pd
from seeq.spy._status import Status
from seeq.spy.workbooks import Analysis, Topic


class BuildMode(Enum):
    BUILD = 'BUILD'
    BROCHURE = 'BROCHURE'


class BuildPhase:
    # This is not an Enum because we want to put strings in the Status DataFrame
    INSTANTIATING = 'Instantiating'
    BUILDING = 'Building'
    SUCCESS = 'Success'


class BuildContext:
    workbooks: dict
    objects: dict
    cache: dict
    results_dict: Dict[tuple, dict]
    push_df: Optional[pd.DataFrame]
    status: Optional[Status]
    mode: BuildMode
    phase: str
    at_least_one_thing_built_somewhere: bool

    def __init__(self, status: Status = None, mode: BuildMode = BuildMode.BUILD, workbooks: list = None):
        self.workbooks = dict()
        self.objects = dict()
        self.cache = dict()
        self.results_dict = dict()
        self.push_df = None
        self.status = status
        self.mode = mode
        self.phase = BuildPhase.INSTANTIATING
        self.at_least_one_thing_built_somewhere = True

        if workbooks is not None:
            self.workbooks = {(w['Workbook Type'], w.name): w for w in workbooks}

            for workbook_type in [Analysis, Topic]:
                workbook_list = [w for w in workbooks if isinstance(w, workbook_type)]
                if len(workbook_list) == 1:
                    # There's only one of this type, so set it as the default for when @Asset.Display or
                    # @Asset.Document decorator does not specify an 'analysis' or 'topic' argument
                    self.workbooks[(workbook_list[0]['Workbook Type'], None)] = workbook_list[0]

    def __repr__(self):
        return f'Mode: {self.mode}, Phase: {self.phase}, Objects: {len(self.objects)}'

    def add_results(self, results_to_add):
        if results_to_add is None:
            return

        if isinstance(results_to_add, list):
            for result_to_add in results_to_add:
                self.results_dict[
                    (result_to_add['Path'], result_to_add['Asset'], result_to_add['Name'])] = result_to_add
        else:
            self.results_dict[
                (results_to_add['Path'], results_to_add['Asset'], results_to_add['Name'])] = results_to_add

    def get_results(self):
        return list(self.results_dict.values())

    def add_object(self, object_to_add):
        path = '' if pd.isna(object_to_add['Path']) else object_to_add['Path']
        key = (path, object_to_add['Asset'], object_to_add.__class__)
        if key in self.objects:
            raise SPyInstanceAlreadyExists(self.objects[key])
        self.objects[key] = object_to_add

    def get_object(self, row, template):
        path = '' if pd.isna(row['Build Path']) else row['Build Path']
        key = (path, row['Build Asset'], template)
        return self.objects[key]


class SPyInstanceAlreadyExists(BaseException):
    def __init__(self, instance):
        self.instance = instance
